from typing import Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.data import Bounded, Composite, Unbounded

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.envs.common.utils import batch_to_scalar
from rl4co.utils.ops import gather_by_index, get_distance, get_tour_length

from .generator import MTSPGenerator
from .render import render


class MTSPEnv(RL4COEnvBase):
    """Multiple Traveling Salesman Problem environment
    At each step, an agent chooses to visit a city. A maximum of `num_agents` agents can be employed to visit the cities.
    The cost can be defined in two ways:
        - `minmax`: (default) the reward is the maximum of the path lengths of all the agents
        - `sum`: the cost is the sum of the path lengths of all the agents
    Reward is - cost, so the goal is to maximize the reward (minimize the cost).

    Observations:
        - locations of the depot and each customer.
        - number of agents.
        - the current agent index.
        - the current location of the vehicle.

    Constrains:
        - each agent's tour starts and ends at the depot.
        - each customer must be visited exactly once.

    Finish condition:
        - all customers are visited and all agents back to the depot.

    Reward:
        There are two ways to calculate the cost (-reward):
        - `minmax`: (default) the cost is the maximum of the path lengths of all the agents.
        - `sum`: the cost is the sum of the path lengths of all the agents.

    Args:
        cost_type: type of cost to use, either `minmax` or `sum`
        generator: MTSPGenerator instance as the data generator
        generator_params: parameters for the generator
    """

    name = "mtsp"

    def __init__(
        self,
        generator: MTSPGenerator = None,
        generator_params: dict = {},
        cost_type: str = "minmax",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if generator is None:
            generator = MTSPGenerator(**generator_params)
        self.generator = generator
        self.cost_type = cost_type
        self._make_spec(self.generator)

    @staticmethod
    def _step(td: TensorDict) -> TensorDict:
        # Initial variables
        is_first_action = batch_to_scalar(td["i"]) == 0
        current_node = td["action"]
        first_node = current_node if is_first_action else td["first_node"]

        # Get the locations of the current node and the previous node and the depot
        cur_loc = gather_by_index(td["locs"], current_node)
        prev_loc = gather_by_index(
            td["locs"], td["current_node"]
        )  # current_node is the previous node
        depot_loc = td["locs"][..., 0, :]

        # If current_node is the depot, then increment agent_idx
        cur_agent_idx = td["agent_idx"] + (current_node == 0).long()

        # Set not visited to 0 (i.e., we visited the node)
        available = td["action_mask"].scatter(
            -1, current_node[..., None].expand_as(td["action_mask"]), 0
        )
        # Available[..., 0] is the depot, which is always available unless:
        # - current_node is the depot
        # - agent_idx greater than num_agents -1
        available[..., 0] = torch.logical_and(
            current_node != 0, td["agent_idx"] < td["num_agents"] - 1
        )

        # We are done there are no unvisited locations except the depot
        done = torch.count_nonzero(available[..., 1:], dim=-1) == 0

        # If done is True, then we make the depot available again, so that it will be selected as the next node with prob 1
        available[..., 0] = torch.logical_or(done, available[..., 0])

        # Update the current length
        current_length = td["current_length"] + get_distance(cur_loc, prev_loc)

        # If done, we add the distance from the current_node to the depot as well
        current_length = torch.where(
            done, current_length + get_distance(cur_loc, depot_loc), current_length
        )

        # We update the max_subtour_length and reset the current_length
        max_subtour_length = torch.where(
            current_length > td["max_subtour_length"],
            current_length,
            td["max_subtour_length"],
        )

        # If current agent is different from previous agent, then we have a new subtour and reset the length
        current_length *= (cur_agent_idx == td["agent_idx"]).float()

        # The reward is the negative of the max_subtour_length (minmax objective)
        reward = -max_subtour_length

        td.update(
            {
                "max_subtour_length": max_subtour_length,
                "current_length": current_length,
                "agent_idx": cur_agent_idx,
                "first_node": first_node,
                "current_node": current_node,
                "i": td["i"] + 1,
                "action_mask": available,
                "reward": reward,
                "done": done,
            }
        )

        return td

    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        device = td.device

        # Keep track of the agent number to know when to stop
        agent_idx = torch.zeros((*batch_size,), dtype=torch.int64, device=device)

        # Make variable for max_subtour_length between subtours
        max_subtour_length = torch.zeros(batch_size, dtype=torch.float32, device=device)
        current_length = torch.zeros(batch_size, dtype=torch.float32, device=device)

        # Other variables
        current_node = torch.zeros((*batch_size,), dtype=torch.int64, device=device)
        available = torch.ones(
            (*batch_size, self.generator.num_loc), dtype=torch.bool, device=device
        )  # 1 means not visited, i.e. action is allowed
        available[..., 0] = 0  # Depot is not available as first node
        i = torch.zeros((*batch_size,), dtype=torch.int64, device=device)

        return TensorDict(
            {
                "locs": td["locs"],  # depot is first node
                "num_agents": td["num_agents"],
                "max_subtour_length": max_subtour_length,
                "current_length": current_length,
                "agent_idx": agent_idx,
                "first_node": current_node,
                "current_node": current_node,
                "i": i,
                "action_mask": available,
            },
            batch_size=batch_size,
        )

    def _make_spec(self, generator: MTSPGenerator):
        """Make the observation and action specs from the parameters."""
        self.observation_spec = Composite(
            locs=Bounded(
                low=generator.min_loc,
                high=generator.max_loc,
                shape=(generator.num_loc, 2),
                dtype=torch.float32,
            ),
            num_agents=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            agent_idx=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            current_length=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            max_subtour_length=Unbounded(
                shape=(1),
                dtype=torch.float32,
            ),
            first_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            current_node=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            i=Unbounded(
                shape=(1),
                dtype=torch.int64,
            ),
            action_mask=Unbounded(
                shape=(generator.num_loc),
                dtype=torch.bool,
            ),
            shape=(),
        )
        self.action_spec = Bounded(
            shape=(1,),
            dtype=torch.int64,
            low=0,
            high=generator.num_loc,
        )
        self.reward_spec = Unbounded()
        self.done_spec = Unbounded(dtype=torch.bool)

    def _get_reward(self, td, actions=None) -> TensorDict:
        # With minmax, get the maximum distance among subtours, calculated in the model
        if self.cost_type == "minmax":
            return td["reward"].squeeze(-1)

        # With distance, same as TSP
        elif self.cost_type == "sum":
            locs = td["locs"]
            locs_ordered = locs.gather(1, actions.unsqueeze(-1).expand_as(locs))
            return -get_tour_length(locs_ordered)

        else:
            raise ValueError(f"Cost type {self.cost_type} not supported")

    @staticmethod
    def check_solution_validity(td: TensorDict, actions: torch.Tensor):
        assert True, "Not implemented"

    @staticmethod
    def render(td, actions=None, ax=None):
        return render(td, actions, ax)



import math

from typing import Tuple

import torch
import torch.nn as nn

from rl4co.utils.ops import gather_by_index


class SkipConnection(nn.Module):
    def __init__(self, module):
        super(SkipConnection, self).__init__()
        self.module = module

    def forward(self, x):
        return x + self.module(x)


class AdaptiveSequential(nn.Sequential):
    def forward(
        self, *inputs: Tuple[torch.Tensor] | torch.Tensor
    ) -> Tuple[torch.Tensor] | torch.Tensor:
        for module in self._modules.values():
            if type(inputs) == tuple:
                inputs = module(*inputs)
            else:
                inputs = module(inputs)
        return inputs


class Normalization(nn.Module):
    def __init__(self, embed_dim, normalization="batch"):
        super(Normalization, self).__init__()
        if normalization != "layer":
            normalizer_class = {
                "batch": nn.BatchNorm1d,
                "instance": nn.InstanceNorm1d,
            }.get(normalization, None)

            self.normalizer = normalizer_class(embed_dim, affine=True)
        else:
            self.normalizer = "layer"

    def forward(self, x):
        if isinstance(self.normalizer, nn.BatchNorm1d):
            return self.normalizer(x.view(-1, x.size(-1))).view(*x.size())
        elif isinstance(self.normalizer, nn.InstanceNorm1d):
            return self.normalizer(x.permute(0, 2, 1)).permute(0, 2, 1)
        elif self.normalizer == "layer":
            return (x - x.mean((1, 2)).view(-1, 1, 1)) / torch.sqrt(
                x.var((1, 2)).view(-1, 1, 1) + 1e-05
            )
        else:
            assert self.normalizer is None, "Unknown normalizer type"
            return x


class PositionalEncoding(nn.Module):
    def __init__(self, embed_dim: int, dropout: float = 0.1, max_len: int = 1000):
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)
        self.d_model = embed_dim
        max_len = max_len
        position = torch.arange(max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, self.d_model, 2) * (-math.log(10000.0) / self.d_model)
        )
        pe = torch.zeros(max_len, 1, self.d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        pe = pe.transpose(0, 1)  # [1, max_len, d_model]
        self.register_buffer("pe", pe)

    def forward(self, hidden: torch.Tensor, seq_pos) -> torch.Tensor:
        """
        Arguments:
            x: Tensor, shape ``[batch_size, seq_len, embedding_dim]``
            seq_pos: Tensor, shape ``[batch_size, seq_len]``
        """
        pes = self.pe.expand(hidden.size(0), -1, -1).gather(
            1, seq_pos.unsqueeze(-1).expand(-1, -1, self.d_model)
        )
        hidden = hidden + pes
        return self.dropout(hidden)


class TransformerFFN(nn.Module):
    def __init__(self, embed_dim, feed_forward_hidden, normalization="batch") -> None:
        super().__init__()

        self.ops = nn.ModuleDict(
            {
                "norm1": Normalization(embed_dim, normalization),
                "ffn": nn.Sequential(
                    nn.Linear(embed_dim, feed_forward_hidden),
                    nn.ReLU(),
                    nn.Linear(feed_forward_hidden, embed_dim),
                ),
                "norm2": Normalization(embed_dim, normalization),
            }
        )

    def forward(self, x, x_old):
        x = self.ops["norm1"](x_old + x)
        x = self.ops["norm2"](x + self.ops["ffn"](x))

        return x


class RandomEncoding(nn.Module):
    """This is like torch.nn.Embedding but with rows of embeddings are randomly
    permuted in each forward pass before lookup operation. This might be useful
    in cases where classes have no fixed meaning but rather indicate a connection
    between different elements in a sequence. Reference is the MatNet model.
    """

    def __init__(self, embed_dim: int, max_classes: int = 100):
        super().__init__()
        self.embed_dim = embed_dim
        self.max_classes = max_classes
        rand_emb = torch.rand(max_classes, self.embed_dim)
        self.register_buffer("emb", rand_emb)

    def forward(self, hidden: torch.Tensor, classes=None) -> torch.Tensor:
        b, s, _ = hidden.shape
        if classes is None:
            classes = torch.eye(s).unsqueeze(0).expand(b, s)
        assert (
            classes.max() < self.max_classes
        ), "number of classes larger than embedding table"
        classes = classes.unsqueeze(-1).expand(-1, -1, self.embed_dim)
        rand_idx = torch.rand(b, self.max_classes).argsort(dim=1)
        embs_permuted = self.emb[rand_idx]
        rand_emb = gather_by_index(embs_permuted, classes, dim=1)
        hidden = hidden + rand_emb
        return hidden



import math

from typing import Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from rl4co.models.common import ImprovementEncoder
from rl4co.models.nn.ops import AdaptiveSequential, Normalization
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


# implements the Multi-head DAC-Att module
class DAC_ATT(nn.Module):
    def __init__(self, n_heads, input_dim, embed_dim=None, val_dim=None, key_dim=None):
        super(DAC_ATT, self).__init__()

        self.n_heads = n_heads

        self.key_dim = self.val_dim = embed_dim // n_heads
        self.input_dim = input_dim
        self.embed_dim = embed_dim

        self.norm_factor = 1 / math.sqrt(1 * self.key_dim)

        # W_h^Q in the paper
        self.W_query_node = nn.Parameter(
            torch.Tensor(n_heads, self.input_dim, self.key_dim)
        )
        # W_g^Q in the paper
        self.W_query_pos = nn.Parameter(
            torch.Tensor(n_heads, self.input_dim, self.key_dim)
        )
        # W_h^K in the paper
        self.W_key_node = nn.Parameter(
            torch.Tensor(n_heads, self.input_dim, self.key_dim)
        )
        # W_g^K in the paper
        self.W_key_pos = nn.Parameter(torch.Tensor(n_heads, self.input_dim, self.key_dim))

        # W_h^V and W_h^Vref in the paper
        self.W_val_node = nn.Parameter(
            torch.Tensor(2 * n_heads, self.input_dim, self.val_dim)
        )
        # W_g^V and W_g^Vref in the paper
        self.W_val_pos = nn.Parameter(
            torch.Tensor(2 * n_heads, self.input_dim, self.val_dim)
        )

        # W_h^O and W_g^O in the paper
        if embed_dim is not None:
            self.W_out_node = nn.Parameter(
                torch.Tensor(n_heads, 2 * self.key_dim, embed_dim)
            )
            self.W_out_pos = nn.Parameter(
                torch.Tensor(n_heads, 2 * self.key_dim, embed_dim)
            )

        self.init_parameters()

    def init_parameters(self):
        for param in self.parameters():
            stdv = 1.0 / math.sqrt(param.size(-1))
            param.data.uniform_(-stdv, stdv)

    def forward(self, h_node_in, h_pos_in):  # input (NFEs, PFEs)
        # h,g should be (batch_size, graph_size, input_dim)
        batch_size, graph_size, input_dim = h_node_in.size()

        shp = (self.n_heads, batch_size, graph_size, -1)
        shp_v = (2, self.n_heads, batch_size, graph_size, -1)

        h_node = h_node_in.contiguous().view(-1, input_dim)
        h_pos = h_pos_in.contiguous().view(-1, input_dim)

        Q_node = torch.matmul(h_node, self.W_query_node).view(shp)
        Q_pos = torch.matmul(h_pos, self.W_query_pos).view(shp)

        K_node = torch.matmul(h_node, self.W_key_node).view(shp)
        K_pos = torch.matmul(h_pos, self.W_key_pos).view(shp)

        V_node = torch.matmul(h_node, self.W_val_node).view(shp_v)
        V_pos = torch.matmul(h_pos, self.W_val_pos).view(shp_v)

        # Get attention correlations and norm by softmax
        node_correlations = self.norm_factor * torch.matmul(
            Q_node, K_node.transpose(2, 3)
        )
        pos_correlations = self.norm_factor * torch.matmul(Q_pos, K_pos.transpose(2, 3))
        attn1 = F.softmax(node_correlations, dim=-1)  # head, bs, n, n
        attn2 = F.softmax(pos_correlations, dim=-1)  # head, bs, n, n

        heads_node_1 = torch.matmul(attn1, V_node[0])  # self-attn
        heads_node_2 = torch.matmul(attn2, V_node[1])  # cross-aspect ref attn

        heads_pos_1 = torch.matmul(attn1, V_pos[0])  # cross-aspect ref attn
        heads_pos_2 = torch.matmul(attn2, V_pos[1])  # self-attn

        heads_node = torch.cat((heads_node_1, heads_node_2), -1)
        heads_pos = torch.cat((heads_pos_1, heads_pos_2), -1)

        # get output
        out_node = torch.mm(
            heads_node.permute(1, 2, 0, 3)
            .contiguous()
            .view(-1, self.n_heads * 2 * self.val_dim),
            self.W_out_node.view(-1, self.embed_dim),
        ).view(batch_size, graph_size, self.embed_dim)

        out_pos = torch.mm(
            heads_pos.permute(1, 2, 0, 3)
            .contiguous()
            .view(-1, self.n_heads * 2 * self.val_dim),
            self.W_out_pos.view(-1, self.embed_dim),
        ).view(batch_size, graph_size, self.embed_dim)

        return out_node, out_pos  # dual-aspect representation (NFEs, PFEs)


# implements the DAC encoder
class DACTEncoderLayer(nn.Module):
    def __init__(
        self,
        n_heads,
        embed_dim,
        feed_forward_hidden,
        normalization="layer",
    ):
        super(DACTEncoderLayer, self).__init__()

        self.MHA_sublayer = DACsubLayer(
            n_heads,
            embed_dim,
            feed_forward_hidden,
            normalization=normalization,
        )

        self.FFandNorm_sublayer = FFNsubLayer(
            n_heads,
            embed_dim,
            feed_forward_hidden,
            normalization=normalization,
        )

    def forward(self, input1, input2):
        out1, out2 = self.MHA_sublayer(input1, input2)
        return self.FFandNorm_sublayer(out1, out2)


# implements the DAC encoder (DAC-Att sublayer)
class DACsubLayer(nn.Module):
    def __init__(
        self,
        n_heads,
        embed_dim,
        feed_forward_hidden,
        normalization="layer",
    ):
        super(DACsubLayer, self).__init__()

        self.MHA = DAC_ATT(n_heads, input_dim=embed_dim, embed_dim=embed_dim)

        self.Norm = Normalization(embed_dim, normalization)

    def forward(self, input1, input2):
        # Attention and Residual connection
        out1, out2 = self.MHA(input1, input2)

        # Normalization
        return self.Norm(out1 + input1), self.Norm(out2 + input2)


# implements the DAC encoder (FFN sublayer)
class FFNsubLayer(nn.Module):
    def __init__(
        self,
        n_heads,
        embed_dim,
        feed_forward_hidden,
        normalization="layer",
    ):
        super(FFNsubLayer, self).__init__()

        self.FF1 = (
            nn.Sequential(
                nn.Linear(embed_dim, feed_forward_hidden),
                nn.ReLU(inplace=True),
                nn.Linear(feed_forward_hidden, embed_dim),
            )
            if feed_forward_hidden > 0
            else nn.Linear(embed_dim, embed_dim)
        )

        self.FF2 = (
            nn.Sequential(
                nn.Linear(embed_dim, feed_forward_hidden),
                nn.ReLU(inplace=True),
                nn.Linear(feed_forward_hidden, embed_dim),
            )
            if feed_forward_hidden > 0
            else nn.Linear(embed_dim, embed_dim)
        )

        self.Norm = Normalization(embed_dim, normalization)

    def forward(self, input1, input2):
        # FF and Residual connection
        out1 = self.FF1(input1)
        out2 = self.FF2(input2)

        # Normalization
        return self.Norm(out1 + input1), self.Norm(out2 + input2)


class DACTEncoder(ImprovementEncoder):
    """Dual-Aspect Collaborative Transformer Encoder as in Ma et al. (2021)

    Args:
        embed_dim: Dimension of the embedding space
        init_embedding: Module to use for the initialization of the node embeddings
        pos_embedding: Module to use for the initialization of the positional embeddings
        env_name: Name of the environment used to initialize embeddings
        pos_type: Name of the used positional encoding method (CPE or APE)
        num_heads: Number of heads in the attention layers
        num_layers: Number of layers in the attention network
        normalization: Normalization type in the attention layers
        feedforward_hidden: Hidden dimension in the feedforward layers
    """

    def __init__(
        self,
        embed_dim: int = 64,
        init_embedding: nn.Module = None,
        pos_embedding: nn.Module = None,
        env_name: str = "tsp_kopt",
        pos_type: str = "CPE",
        num_heads: int = 4,
        num_layers: int = 3,
        normalization: str = "layer",
        feedforward_hidden: int = 64,
    ):
        super(DACTEncoder, self).__init__(
            embed_dim=embed_dim,
            env_name=env_name,
            pos_type=pos_type,
            num_heads=num_heads,
            num_layers=num_layers,
            normalization=normalization,
            feedforward_hidden=feedforward_hidden,
        )

        assert self.env_name in ["tsp_kopt"], NotImplementedError()

        self.net = AdaptiveSequential(
            *(
                DACTEncoderLayer(
                    num_heads,
                    embed_dim,
                    feedforward_hidden,
                    normalization,
                )
                for _ in range(num_layers)
            )
        )

    def _encoder_forward(self, init_h: Tensor, init_p: Tensor) -> Tuple[Tensor, Tensor]:
        NFE, PFE = self.net(init_h, init_p)

        return NFE, PFE



import torch

from einops import einsum, reduce
from tensordict import TensorDict
from torch._tensor import Tensor

from rl4co.envs.scheduling.fjsp.env import FJSPEnv
from rl4co.utils.ops import gather_by_index

from .generator import JSSPFileGenerator, JSSPGenerator


class JSSPEnv(FJSPEnv):
    """Job-Shop Scheduling Problem (JSSP) environment
    At each step, the agent chooses a job. The operation to be processed next for the selected job is
    then executed on the associated machine. The reward is 0 unless the agent scheduled all operations of all jobs.
    In that case, the reward is (-)makespan of the schedule: maximizing the reward is equivalent to minimizing the makespan.
    NOTE: The JSSP is a special case of the FJSP, when the number of eligible machines per operation is equal to one for all
    operations. Therefore, this environment is a subclass of the FJSP environment.
    Observations:
        - time: current time
        - next_op: next operation per job
        - proc_times: processing time of operation-machine pairs
        - pad_mask: specifies padded operations
        - start_op_per_job: id of first operation per job
        - end_op_per_job: id of last operation per job
        - start_times: start time of operation (defaults to 0 if not scheduled)
        - finish_times: finish time of operation (defaults to INIT_FINISH if not scheduled)
        - job_ops_adj: adjacency matrix specifying job-operation affiliation
        - ops_job_map: same as above but using ids of jobs to indicate affiliation
        - ops_sequence_order: specifies the order in which operations have to be processed
        - ma_assignment: specifies which operation has been scheduled on which machine
        - busy_until: specifies until when the machine will be busy
        - num_eligible: number of machines that can process an operation
        - job_in_process: whether job is currently being processed
        - job_done: whether the job is done

    Constrains:
        the agent may not select:
        - jobs that are done already
        - jobs that are currently processed

    Finish condition:
        - the agent has scheduled all operations of all jobs

    Reward:
        - the negative makespan of the final schedule

    Args:
        generator: JSSPGenerator instance as the data generator
        generator_params: parameters for the generator
        mask_no_ops: if True, agent may not select waiting operation (unless instance is done)
    """

    name = "jssp"

    def __init__(
        self,
        generator: JSSPGenerator = None,
        generator_params: dict = {},
        mask_no_ops: bool = True,
        **kwargs,
    ):
        if generator is None:
            if generator_params.get("file_path", None) is not None:
                generator = JSSPFileGenerator(**generator_params)
            else:
                generator = JSSPGenerator(**generator_params)

        super().__init__(generator, generator_params, mask_no_ops, **kwargs)

    def _get_features(self, td):
        td = super()._get_features(td)
        # get the id of the machine that executes an operation:
        # (bs, ops, ma)
        ops_ma_adj = td["ops_ma_adj"].transpose(1, 2)
        # (bs, jobs, ma)
        ma_of_next_op = gather_by_index(ops_ma_adj, td["next_op"], dim=1)
        # (bs, jobs)
        td["next_ma"] = ma_of_next_op.argmax(-1)

        # adjacency matrix specifying neighbors of an operation, including its
        # predecessor and successor operations and operations on the same machine
        ops_on_same_ma_adj = einsum(
            td["ops_ma_adj"], td["ops_ma_adj"], "b m o1, b m o2 -> b o1 o2 "
        )
        # concat pred, succ and ops on same machine
        adj = torch.cat((td["ops_adj"], ops_on_same_ma_adj.unsqueeze(-1)), dim=-1).sum(-1)
        # mask padded operations and those scheduled
        mask = td["pad_mask"] + td["op_scheduled"]
        adj.masked_fill_(mask.unsqueeze(1), 0)
        td["adjacency"] = adj

        return td

    def get_action_mask(self, td: TensorDict) -> Tensor:
        action_mask = self._get_job_machine_availability(td)
        if self.mask_no_ops:
            # masking is only allowed if instance is finished
            no_op_mask = td["done"]
        else:
            # if no job is currently processed and instance is not finished yet, waiting is not allowed
            no_op_mask = (
                td["job_in_process"].any(1, keepdims=True) & (~td["done"])
            ) | td["done"]
        # reduce action mask to correspond with logit shape
        action_mask = reduce(action_mask, "bs j m -> bs j", reduction="all")
        # NOTE: 1 means feasible action, 0 means infeasible action
        # (bs, 1 + n_j)
        mask = torch.cat((no_op_mask, ~action_mask), dim=1)
        return mask

    def _translate_action(self, td):
        job = td["action"]
        op = gather_by_index(td["next_op"], job, dim=1)
        # get the machine that corresponds to the selected operation
        ma = gather_by_index(td["ops_ma_adj"], op.unsqueeze(1), dim=2).nonzero()[:, 1]
        return job, op, ma

    @staticmethod
    def load_data(fpath, batch_size=[]):
        g = JSSPFileGenerator(fpath)
        return g(batch_size=batch_size)



from functools import partial

import torch

from torch.utils.data import DataLoader

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl import REINFORCE
from rl4co.models.rl.reinforce.baselines import (
    REINFORCEBaseline,
    RolloutBaseline,
    WarmupBaseline,
)
from rl4co.models.zoo.mdam.policy import MDAMPolicy


def rollout(self, model, env, batch_size=64, device="cpu", dataset=None):
    """In this case the reward from the model is [batch, num_paths]
    and the baseline takes the maximum reward from the model as the baseline.
    https://github.com/liangxinedu/MDAM/blob/19b0bf813fb2dbec2fcde9e22eb50e04675400cd/train.py#L38C29-L38C33
    """
    # if dataset is None, use the dataset of the baseline
    dataset = self.dataset if dataset is None else dataset

    model.eval()
    model = model.to(device)

    def eval_model(batch):
        with torch.inference_mode():
            batch = env.reset(batch.to(device))
            return model(batch, env, decode_type="greedy")["reward"].max(1).values

    dl = DataLoader(dataset, batch_size=batch_size, collate_fn=dataset.collate_fn)

    rewards = torch.cat([eval_model(batch) for batch in dl], 0)
    return rewards


class MDAM(REINFORCE):
    """Multi-Decoder Attention Model (MDAM) is a model
    to train multiple diverse policies, which effectively increases the chance of finding
    good solutions compared with existing methods that train only one policy.
    Reference link: https://arxiv.org/abs/2012.10638;
    Implementation reference: https://github.com/liangxinedu/MDAM.

    Args:
        env: Environment to use for the algorithm
        policy: Policy to use for the algorithm
        baseline: REINFORCE baseline. Defaults to rollout (1 epoch of exponential, then greedy rollout baseline)
        policy_kwargs: Keyword arguments for policy
        baseline_kwargs: Keyword arguments for baseline
        **kwargs: Keyword arguments passed to the superclass
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: MDAMPolicy = None,
        baseline: REINFORCEBaseline | str = "rollout",
        policy_kwargs={},
        baseline_kwargs={},
        **kwargs,
    ):
        if policy is None:
            policy = MDAMPolicy(env_name=env.name, **policy_kwargs)

        super().__init__(env, policy, baseline, baseline_kwargs, **kwargs)

        # Change rollout of baseline to the rollout function
        if isinstance(self.baseline, WarmupBaseline):
            if isinstance(self.baseline.baseline, RolloutBaseline):
                self.baseline.baseline.rollout = partial(rollout, self.baseline.baseline)
        elif isinstance(self.baseline, RolloutBaseline):
            self.baseline.rollout = partial(rollout, self.baseline)

    def calculate_loss(
        self,
        td,
        batch,
        policy_out,
        reward=None,
        log_likelihood=None,
    ):
        """Calculate loss for REINFORCE algorithm.
        Same as in :class:`REINFORCE`, but the bl_val is calculated is simply unsqueezed to match
        the reward shape (i.e., [batch, num_paths])

        Args:
            td: TensorDict containing the current state of the environment
            batch: Batch of data. This is used to get the extra loss terms, e.g., REINFORCE baseline
            policy_out: Output of the policy network
            reward: Reward tensor. If None, it is taken from `policy_out`
            log_likelihood: Log-likelihood tensor. If None, it is taken from `policy_out`
        """
        # Extra: this is used for additional loss terms, e.g., REINFORCE baseline
        extra = batch.get("extra", None)
        reward = reward if reward is not None else policy_out["reward"]
        log_likelihood = (
            log_likelihood if log_likelihood is not None else policy_out["log_likelihood"]
        )

        # REINFORCE baseline
        bl_val, bl_loss = (
            self.baseline.eval(td, reward, self.env) if extra is None else (extra, 0)
        )

        # Main loss function
        # reward: [batch, num_paths]. Note that the baseline value is the max reward
        # if bl_val is a tensor, unsqueeze it to match the reward shape
        if isinstance(bl_val, torch.Tensor):
            if len(bl_val.shape) > 0:
                bl_val = bl_val.unsqueeze(1)
        advantage = reward - bl_val  # advantage = reward - baseline
        reinforce_loss = -(advantage * log_likelihood).mean()
        loss = reinforce_loss + bl_loss
        policy_out.update(
            {
                "loss": loss,
                "reinforce_loss": reinforce_loss,
                "bl_loss": bl_loss,
                "bl_val": bl_val,
            }
        )
        return policy_out



import abc

from typing import Any, Optional

from lightning.pytorch.utilities.types import STEP_OUTPUT
from torch.utils.data import Dataset

from rl4co.models.rl.common.base import RL4COLitModule


class TransductiveModel(RL4COLitModule, metaclass=abc.ABCMeta):
    """Base class for transductive algorithms (i.e. that optimize policy parameters for
    specific instances, see https://en.wikipedia.org/wiki/Transduction_(machine_learning)).
    Transductive algorithms are used online to find better solutions for a given dataset, i.e.
    given a policy, improve (a part of) its parameters such that
    the policy performs better on the given dataset.

    Note:
        By default, we use manual optimization to handle the search.

    Args:
        env: RL4CO environment
        policy: policy network
        dataset: dataset to use for training
        batch_size: batch size
        max_iters: maximum number of iterations
        max_runtime: maximum runtime in seconds
        save_path: path to save the model
        **kwargs: additional arguments
    """

    def __init__(
        self,
        env,
        policy,
        dataset: Dataset | str,
        batch_size: int = 1,
        max_iters: int = 100,
        max_runtime: Optional[int] = 86_400,
        save_path: Optional[str] = None,
        **kwargs,
    ):
        self.save_hyperparameters(logger=False, ignore=["env", "policy", "dataset"])
        super().__init__(env, policy, **kwargs)
        self.dataset = dataset
        self.automatic_optimization = False  # we optimize manually

    def setup(self, stage="fit"):
        """Setup the dataset and attributes.
        The RL4COLitModulebase class automatically loads the data.
        """
        if isinstance(self.dataset, str):
            # load from file
            self.dataset = self.env.dataset(filename=self.dataset)

        # Set all datasets and batch size as the same
        for split in ["train", "val", "test"]:
            setattr(self, f"{split}_dataset", self.dataset)
            setattr(self, f"{split}_batch_size", self.hparams.batch_size)

        # Setup loggers
        self.setup_loggers()

    def on_train_batch_start(self, batch: Any, batch_idx: int):
        """Called before training (i.e. search) for a new batch begins.
        This can be used to perform changes to the model or optimizer at the start of each batch.
        """
        pass  # Implement in subclass

    @abc.abstractmethod
    def training_step(self, batch, batch_idx):
        """Main search loop. We use the training step to effectively adapt to a `batch` of instances."""
        raise NotImplementedError("Implement in subclass")

    def on_train_batch_end(
        self, outputs: STEP_OUTPUT, batch: Any, batch_idx: int
    ) -> None:
        """Called when the train batch ends. This can be used for
        instance for logging or clearing cache.
        """
        pass  # Implement in subclass

    def on_train_epoch_end(self) -> None:
        """Called when the train ends."""
        pass  # Implement in subclass

    def validation_step(self, batch: Any, batch_idx: int):
        """Not used during search"""
        pass

    def test_step(self, batch: Any, batch_idx: int):
        """Not used during search"""
        pass



from typing import Tuple

import torch.nn as nn

from tensordict import TensorDict
from torch import Tensor

from rl4co.envs import RL4COEnvBase
from rl4co.models.common.constructive import AutoregressiveEncoder
from rl4co.models.nn.env_embeddings import env_init_embedding
from rl4co.models.nn.graph.attnnet import GraphAttentionNetwork


class AttentionModelEncoder(AutoregressiveEncoder):
    """Graph Attention Encoder as in Kool et al. (2019).
    First embed the input and then process it with a Graph Attention Network.

    Args:
        embed_dim: Dimension of the embedding space
        init_embedding: Module to use for the initialization of the embeddings
        env_name: Name of the environment used to initialize embeddings
        num_heads: Number of heads in the attention layers
        num_layers: Number of layers in the attention network
        normalization: Normalization type in the attention layers
        feedforward_hidden: Hidden dimension in the feedforward layers
        net: Graph Attention Network to use
        sdpa_fn: Function to use for the scaled dot product attention
        moe_kwargs: Keyword arguments for MoE
    """

    def __init__(
        self,
        embed_dim: int = 128,
        init_embedding: nn.Module = None,
        env_name: str = "tsp",
        num_heads: int = 8,
        num_layers: int = 3,
        normalization: str = "batch",
        feedforward_hidden: int = 512,
        net: nn.Module = None,
        sdpa_fn=None,
        moe_kwargs: dict = None,
    ):
        super(AttentionModelEncoder, self).__init__()

        if isinstance(env_name, RL4COEnvBase):
            env_name = env_name.name
        self.env_name = env_name

        self.init_embedding = (
            env_init_embedding(self.env_name, {"embed_dim": embed_dim})
            if init_embedding is None
            else init_embedding
        )

        self.net = (
            GraphAttentionNetwork(
                num_heads,
                embed_dim,
                num_layers,
                normalization,
                feedforward_hidden,
                sdpa_fn=sdpa_fn,
                moe_kwargs=moe_kwargs,
            )
            if net is None
            else net
        )

    def forward(
        self, td: TensorDict, mask: Tensor | None = None
    ) -> Tuple[Tensor, Tensor]:
        """Forward pass of the encoder.
        Transform the input TensorDict into a latent representation.

        Args:
            td: Input TensorDict containing the environment state
            mask: Mask to apply to the attention

        Returns:
            h: Latent representation of the input
            init_h: Initial embedding of the input
        """
        # Transfer to embedding space
        init_h = self.init_embedding(td)

        # Process embedding
        h = self.net(init_h, mask)

        # Return latent representation and initial embedding
        return h, init_h



from functools import lru_cache, cached_property
from typing import Optional, Tuple

import torch

from tensordict import TensorDict
from torch import Tensor
from tqdm import trange

from rl4co.envs import RL4COEnvBase
from rl4co.models.common.constructive.nonautoregressive.decoder import (
    NonAutoregressiveDecoder,
)
from rl4co.utils.decoding import Sampling
from rl4co.utils.ops import batchify, get_distance_matrix, unbatchify


class AntSystem:
    """Implements the Ant System algorithm: https://doi.org/10.1109/3477.484436.

    Args:
        log_heuristic: Logarithm of the heuristic matrix.
        n_ants: Number of ants to be used in the algorithm. Defaults to 20.
        alpha: Importance of pheromone in the decision-making process. Defaults to 1.0.
        beta: Importance of heuristic information in the decision-making process. Defaults to 1.0.
        decay: Rate at which pheromone evaporates. Should be between 0 and 1. Defaults to 0.95.
        Q: Rate at which pheromone deposits. Defaults to `1 / n_ants`.
        pheromone: Initial pheromone matrix. Defaults to `torch.ones_like(log_heuristic)`.
        use_local_search: Whether to use local_search provided by the env. Default to False.
        use_nls: Whether to use neural-guided local search provided by the env. Default to False.
        n_perturbations: Number of perturbations to be used for nls. Defaults to 5.
        local_search_params: Arguments to be passed to the local_search.
        perturbation_params: Arguments to be passed to the perturbation used for nls.
    """

    def __init__(
        self,
        log_heuristic: Tensor,
        n_ants: int = 20,
        alpha: float = 1.0,
        beta: float = 1.0,
        decay: float = 0.95,
        Q: Optional[float] = None,
        pheromone: Optional[Tensor | int] = None,
        use_local_search: bool = False,
        use_nls: bool = False,
        n_perturbations: int = 1,
        local_search_params: dict = {},
        perturbation_params: dict = {},
    ):
        self.batch_size = log_heuristic.shape[0]
        self.n_ants = n_ants
        self.alpha = alpha
        self.beta = beta
        self.decay = decay
        self.Q = 1 / self.n_ants / self.decay if Q is None else Q

        self.log_heuristic = log_heuristic

        if pheromone is None or isinstance(pheromone, int):
            self.pheromone = torch.ones_like(log_heuristic)
            self.pheromone.fill_(pheromone if isinstance(pheromone, int) else 1)
        else:
            assert pheromone.shape == log_heuristic.shape
            self.pheromone = pheromone

        self.final_actions = self.final_reward = None
        self.final_reward_cache: dict = {}

        self.use_local_search = use_local_search
        assert not (use_nls and not use_local_search), "use_nls requires use_local_search"
        self.use_nls = use_nls
        self.n_perturbations = n_perturbations
        self.local_search_params = local_search_params.copy()  # needs to be copied to avoid side effects
        self.perturbation_params = perturbation_params.copy()

        self._batchindex = torch.arange(self.batch_size, device=log_heuristic.device)

    @cached_property
    def heuristic_dist(self) -> torch.Tensor:
        heuristic = self.log_heuristic.exp().detach().cpu() + 1e-10
        heuristic_dist = 1 / (heuristic / heuristic.max(-1, keepdim=True)[0] + 1e-5)
        heuristic_dist[
            :, torch.arange(heuristic_dist.shape[1]), torch.arange(heuristic_dist.shape[2])
        ] = 0
        return heuristic_dist

    def run(
        self,
        td_initial: TensorDict,
        env: RL4COEnvBase,
        n_iterations: int,
        decoding_kwargs: dict,
        disable_tqdm: bool = True,
    ) -> Tuple[Tensor, dict[int, Tensor]]:
        """Run the Ant System algorithm for a specified number of iterations.

        Args:
            td_initial: Initial state of the problem.
            env: Environment representing the problem.
            n_iterations: Number of iterations to run the algorithm.
            decoding_kwargs: Keyword arguments for decoding strategy.
            disable_tqdm: Whether to disable the tqdm progress bar. Defaults to False.

        Returns:
            td: The final state of the problem.
            actions: The final actions chosen by the algorithm.
            reward: The final reward achieved by the algorithm.
        """
        pbar = trange(
            n_iterations, dynamic_ncols=True, desc="Running ACO", disable=disable_tqdm
        )
        for i in pbar:
            # reset environment
            td = td_initial.clone()
            self._one_step(td, env, decoding_kwargs)
            self.final_reward_cache[i] = self.final_reward.clone()  # type: ignore
            pbar.set_postfix({"reward": f"{self.final_reward.mean().item():.3f}"})  # type: ignore

        action_matrix = self._convert_final_action_to_matrix()
        return action_matrix, self.final_reward_cache

    def _one_step(self, td: TensorDict, env: RL4COEnvBase, decoding_kwargs: dict):
        """Run one step of the Ant System algorithm.

        Args:
            td: Current state of the problem.
            env: Environment representing the problem.
            decoding_kwargs: Keyword arguments for decoding strategy.

        Returns:
            actions: The actions chosen by the algorithm.
            reward: The reward achieved by the algorithm.
        """
        # sampling
        td, env, actions, reward = self._sampling(td, env, decoding_kwargs)
        # local search, reserved for extensions
        if self.use_local_search:
            actions, reward = self.local_search(td, env, actions, decoding_kwargs)

        # reshape from (batch_size * n_ants, ...) to (batch_size, n_ants, ...)
        reward = unbatchify(reward, self.n_ants)
        actions = unbatchify(actions, self.n_ants)

        # update final actions and rewards
        self._update_results(actions, reward)
        # update pheromone matrix
        self._update_pheromone(actions, reward)

        return actions, reward

    def _sampling(
        self,
        td: TensorDict,
        env: RL4COEnvBase,
        decoding_kwargs: dict,
    ):
        # Sample from heatmaps
        # p = phe**alpha * heu**beta <==> log(p) = alpha*log(phe) + beta*log(heu)
        heatmaps_logits = (
            self.alpha * torch.log(self.pheromone) + self.beta * self.log_heuristic
        )
        decode_strategy = Sampling(**decoding_kwargs)

        td, env, num_starts = decode_strategy.pre_decoder_hook(td, env)
        while not td["done"].all():
            logits, mask = NonAutoregressiveDecoder.heatmap_to_logits(
                td, heatmaps_logits, num_starts
            )
            td = decode_strategy.step(logits, mask, td)
            td = env.step(td)["next"]

        logprobs, actions, td, env = decode_strategy.post_decoder_hook(td, env)
        reward = env.get_reward(td, actions)

        return td, env, actions, reward

    def local_search(
        self, td: TensorDict, env: RL4COEnvBase, actions: Tensor, decoding_kwargs: dict
    ) -> Tuple[Tensor, Tensor]:
        """Perform local search on the actions and reward obtained.

        Args:
            td: Current state of the problem.
            env: Environment representing the problem.
            actions: Actions chosen by the algorithm.

        Returns:
            actions: The modified actions
            reward: The modified reward
        """
        device = td.device
        if env.name in ["tsp", "cvrp"]:
            # Convert to CPU in advance to minimize the overhead from device transfer
            td = td.detach().cpu()
            # TODO: avoid or generalize this, e.g., pre-compute for local search in each env
            td["distances"] = get_distance_matrix(td["locs"])
            actions = actions.detach().cpu()
        elif env.name in ["pctsp", "op"]:  # destroy & repair local search
            self.local_search_params.update(
                {
                    "decoding_kwargs": decoding_kwargs,
                    "heatmap": batchify(self.log_heuristic, self.n_ants),
                }
            )
        else:
            raise NotImplementedError(f"Local search not implemented for {env.name}")

        best_actions = env.local_search(td=td, actions=actions, **self.local_search_params)
        best_rewards = env.get_reward(td, best_actions)

        if self.use_nls:
            td_perturb = td.clone()
            td_perturb["distances"] = torch.tile(self.heuristic_dist, (self.n_ants, 1, 1))
            new_actions = best_actions.clone()

            for _ in range(self.n_perturbations):
                perturbed_actions = env.local_search(
                    td=td_perturb, actions=new_actions, **self.perturbation_params
                )
                new_actions = env.local_search(
                    td=td, actions=perturbed_actions, **self.local_search_params
                )
                new_rewards = env.get_reward(td, new_actions)

                improved_indices = new_rewards > best_rewards
                best_actions = env.replace_selected_actions(
                    best_actions, new_actions, improved_indices
                )
                best_rewards[improved_indices] = new_rewards[improved_indices]

        best_actions = best_actions.to(device)
        best_rewards = best_rewards.to(device)

        return best_actions, best_rewards

    def _update_results(self, actions, reward):
        # update the best-trails recorded in self.final_actions
        best_index = reward.argmax(-1)
        best_reward = reward[self._batchindex, best_index]
        best_actions = actions[self._batchindex, best_index]

        if self.final_actions is None or self.final_reward is None:
            self.final_actions = list(iter(best_actions.clone()))
            self.final_reward = best_reward.clone()
        else:
            require_update = self._batchindex[self.final_reward <= best_reward]
            for index in require_update:
                self.final_actions[index] = best_actions[index].clone()
            self.final_reward[require_update] = best_reward[require_update].clone()

        return best_index

    def _update_pheromone(self, actions, reward):
        # calculate Δphe
        delta_pheromone = torch.zeros_like(self.pheromone)
        from_node = actions[:, :, :-1]
        to_node = actions[:, :, 1:]
        mapped_reward = self._reward_map(reward).detach()
        batch_action_indices = self._batch_action_indices(
            self.batch_size, actions.shape[-1] - 1, reward.device
        )

        for ant_index in range(self.n_ants):
            delta_pheromone[
                batch_action_indices,
                from_node[:, ant_index].flatten(),
                to_node[:, ant_index].flatten(),
            ] += mapped_reward[batch_action_indices, ant_index]
        delta_pheromone[:, 0, 0] = 0  # to ignore the tailing zeros

        # decay & update
        self.pheromone *= self.decay
        self.pheromone += delta_pheromone

    def _reward_map(self, x: Tensor):
        """Map reward $f: \\mathbb{R} \\rightarrow \\mathbb{R}^+$"""
        M, _ = x.max(-1, keepdim=True)
        m, _ = x.min(-1, keepdim=True)
        v = ((x - m) / (M - m)) ** 2 * self.Q
        return v

    @staticmethod
    @lru_cache(5)
    def _batch_action_indices(batch_size: int, n_actions: int, device: torch.device):
        batchindex = torch.arange(batch_size, device=device)
        return batchindex.unsqueeze(1).repeat(1, n_actions).view(-1)

    def _convert_final_action_to_matrix(self) -> Tensor:
        assert self.final_actions is not None
        action_count = max(len(actions) for actions in self.final_actions)
        mat_actions = torch.zeros(
            (self.batch_size, action_count),
            device=self.final_actions[0].device,
            dtype=self.final_actions[0].dtype,
        )
        for index, action in enumerate(self.final_actions):
            mat_actions[index, : len(action)] = action
        return mat_actions



from typing import Any

import torch
import torch.nn as nn

from rl4co.envs.common.base import RL4COEnvBase
from rl4co.models.rl.common.base import RL4COLitModule
from rl4co.models.rl.common.critic import CriticNetwork
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class Memory:
    def __init__(self):
        self.tds = []
        self.actions = []
        self.logprobs = []
        self.rewards = []

    def clear_memory(self):
        del self.tds[:]
        del self.actions[:]
        del self.logprobs[:]
        del self.rewards[:]


class n_step_PPO(RL4COLitModule):
    """
    An implementation of the n-step dactProximal Policy Optimization (PPO) algorithm (https://arxiv.org/abs/2110.02544)
    is presented for training improvement models.
    """

    def __init__(
        self,
        env: RL4COEnvBase,
        policy: nn.Module,
        critic: CriticNetwork = None,
        critic_kwargs: dict = {},
        clip_range: float = 0.1,  # epsilon of PPO
        ppo_epochs: int = 3,  # inner epoch, K
        vf_lambda: float = 1.0,  # lambda of Value function fitting
        normalize_adv: bool = False,  # whether to normalize advantage
        max_grad_norm: float = 0.05,  # max gradient norm
        gamma: float = 0.999,  # gamma for improvement MDP task
        n_step: float = 5,  # n-step for n-step PPO
        T_train: int = 250,  # the maximum inference T used for training
        T_test: int = 1000,  # the maximum inference T used for test
        lr_policy: float = 8e-5,  # the learning rate for actor
        lr_critic: float = 2e-5,  # the learning rate for critic
        CL_scalar: float = 2.0,  # hyperparameter of CL scalar of PPO-CL algorithm
        CL_best: bool = False,  # whether use the best solution from the CL rollout
        metrics: dict = {
            "train": ["loss", "surrogate_loss", "value_loss", "cost_bsf", "cost_init"],
            "val": ["cost_bsf", "cost_init"],
            "test": ["cost_bsf", "cost_init"],
        },
        lr_scheduler=torch.optim.lr_scheduler.ExponentialLR,
        lr_scheduler_kwargs: dict = {
            "gamma": 0.985,  # the learning decay per epoch,
        },
        lr_scheduler_interval: str = "epoch",
        lr_scheduler_monitor=None,
        **kwargs,
    ):
        super().__init__(
            env,
            policy,
            metrics=metrics,
            lr_scheduler=lr_scheduler,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            lr_scheduler_interval=lr_scheduler_interval,
            lr_scheduler_monitor=lr_scheduler_monitor,
            **kwargs,
        )

        self.CL_scalar = CL_scalar
        self.CL_num = 0.0
        self.CL_best = CL_best
        self.automatic_optimization = False  # n_step_PPO uses custom optimization routine

        self.critic = critic

        self.ppo_cfg = {
            "clip_range": clip_range,
            "ppo_epochs": ppo_epochs,
            "vf_lambda": vf_lambda,
            "normalize_adv": normalize_adv,
            "max_grad_norm": max_grad_norm,
            "gamma": gamma,
            "n_step": n_step,
            "T_train": T_train,
            "T_test": T_test,
            "lr_policy": lr_policy,
            "lr_critic": lr_critic,
        }

    def configure_optimizers(self):
        parameters = [
            {"params": self.policy.parameters(), "lr": self.ppo_cfg["lr_policy"]}
        ] + [{"params": self.critic.parameters(), "lr": self.ppo_cfg["lr_critic"]}]

        return super().configure_optimizers(parameters)

    def on_train_epoch_end(self):
        """
        Learning rate scheduler and CL scheduler
        """
        # Learning rate scheduler
        sch = self.lr_schedulers()
        sch.step()

        # CL scheduler
        self.CL_num += 1 / self.CL_scalar

    def shared_step(
        self, batch: Any, batch_idx: int, phase: str, dataloader_idx: int = None
    ):
        if phase != "train":
            with torch.no_grad():
                td = self.env.reset(batch)
                cost_init = td["cost_current"]
                for i in range(self.ppo_cfg["T_test"]):
                    out = self.policy(td, self.env, phase=phase)
                    self.env.step(td)
                out["cost_bsf"] = td["cost_bsf"]

        else:
            # init the training
            memory = Memory()
            td = self.env.reset(batch)

            # perform CL strategy
            with torch.no_grad():
                for i in range(int(self.CL_num)):
                    out = self.policy(td, self.env, phase=phase)
                    self.env.step(td)
            if self.CL_best:
                td = self.env.step_to_solution(td, td["rec_best"])
            cost_init = td["cost_current"]

            # perform gradiant updates every n_step untill reaching T_max
            assert (
                self.ppo_cfg["T_train"] % self.ppo_cfg["n_step"] == 0
            ), "T_max should be divided by n_step with no remainder"
            t = 0
            while t < self.ppo_cfg["T_train"]:
                memory.clear_memory()
                bl = []
                ll = []
                # Rollout for n_step,  perform actor and critic and env step, store the information in memory
                for i in range(self.ppo_cfg["n_step"]):
                    memory.tds.append(td.clone())

                    out = self.policy(td, self.env, phase=phase, return_embeds=True)
                    value_pred = self.critic(
                        out["embeds"].detach(), td["cost_bsf"].unsqueeze(-1)
                    )

                    memory.actions.append(out["actions"].clone())
                    memory.logprobs.append(out["log_likelihood"].clone())
                    bl.append(value_pred)

                    self.env.step(td)
                    memory.rewards.append(td["reward"].clone().view(-1, 1))

                t += self.ppo_cfg["n_step"]

                # PPO inner epoch, K
                old_value = None
                for k in range(self.ppo_cfg["ppo_epochs"]):
                    if k == 0:
                        ll = memory.logprobs

                    else:
                        ll = []
                        bl = []
                        for i in range(self.ppo_cfg["n_step"]):
                            out = self.policy(
                                memory.tds[i].clone(),
                                actions=memory.actions[i],
                                env=self.env,
                                phase=phase,
                                return_actions=False,
                                return_embeds=True,
                            )
                            bl_value = self.critic(
                                out["embeds"].detach(),
                                memory.tds[i]["cost_bsf"].unsqueeze(-1),
                            )

                            ll.append(out["log_likelihood"])
                            bl.append(bl_value)

                    # prepare loglikelihood (ll) and baseline value (bl)
                    ll = torch.stack(ll).view(-1, 1)
                    bl = torch.stack(bl).view(-1, 1)
                    old_ll = torch.stack(memory.logprobs).view(-1, 1)

                    # Compute the Reward wrt n_step
                    Reward = []
                    reward_reversed = memory.rewards[::-1]
                    R = self.critic(
                        self.policy(td, self.env, phase=phase, only_return_embed=True)[
                            "embeds"
                        ].detach(),
                        td["cost_bsf"].unsqueeze(-1),
                    ).detach()  # Remember to detach() since we only need the predicted value here
                    for r in range(len(reward_reversed)):
                        R = R * self.ppo_cfg["gamma"] + reward_reversed[r]
                        Reward.append(R.clone())
                    Reward = torch.stack(Reward[::-1]).view(-1, 1)

                    # Compute the ratio of probabilities of new and old actions
                    ratio = torch.exp(ll - old_ll.detach())

                    # Compute the advantage
                    adv = Reward - bl.detach()

                    # Normalize advantage
                    if self.ppo_cfg["normalize_adv"]:
                        adv = (adv - adv.mean()) / (adv.std() + 1e-8)

                    # Compute the surrogate loss
                    surrogate_loss = -torch.min(
                        ratio * adv,
                        torch.clamp(
                            ratio,
                            1 - self.ppo_cfg["clip_range"],
                            1 + self.ppo_cfg["clip_range"],
                        )
                        * adv,
                    ).mean()

                    # compute value function loss
                    if old_value is None:
                        value_loss = ((bl - Reward) ** 2).mean()
                        old_value = bl.detach()
                    else:
                        value_clipped = (
                            torch.clamp(
                                bl - old_value,
                                -self.ppo_cfg["clip_range"],
                                self.ppo_cfg["clip_range"],
                            )
                            + old_value
                        )

                        value_loss = torch.max(
                            (bl - Reward) ** 2,
                            (value_clipped - Reward) ** 2,
                        ).mean()

                    # compute total loss
                    loss = surrogate_loss + self.ppo_cfg["vf_lambda"] * value_loss

                    # perform manual optimization following the Lightning routine
                    # https://lightning.ai/docs/pytorch/stable/common/optimization.html
                    opt = self.optimizers()
                    opt.zero_grad()
                    self.manual_backward(loss)
                    if self.ppo_cfg["max_grad_norm"] is not None:
                        self.clip_gradients(
                            opt,
                            gradient_clip_val=self.ppo_cfg["max_grad_norm"],
                            gradient_clip_algorithm="norm",
                        )
                    opt.step()

            out.update(
                {
                    "cost_init": cost_init,
                    "cost_bsf": td["cost_bsf"],
                    "loss": loss,
                    "surrogate_loss": surrogate_loss,
                    "value_loss": value_loss,
                }
            )
        metrics = self.log_metrics(out, phase, dataloader_idx=dataloader_idx)
        return {"loss": out.get("loss", None), **metrics}



import torch
import torch.nn as nn

try:
    import torch_geometric.nn as gnn
except ImportError:
    gnn = None

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class GNNLayer(nn.Module):
    """Graph Neural Network Layer for processing graph structures.

    Args:
        units: The number of units in each linear transformation layer.
        act_fn: The name of the activation function to use after each linear layer. Defaults to 'silu'.
        agg_fn: The name of the global aggregation function to use for pooling features across the graph. Defaults to 'mean'.
    """

    def __init__(self, units: int, act_fn: str = "silu", agg_fn: str = "mean"):
        assert gnn is not None, (
            "torch_geometric not found. Please install torch_geometric using instructions from "
            "https://pytorch-geometric.readthedocs.io/en/latest/install/installation.html."
        )

        super(GNNLayer, self).__init__()
        self.units = units
        self.act_fn = getattr(nn.functional, act_fn)
        self.agg_fn = getattr(gnn, f"global_{agg_fn}_pool")

        # Vertex updates
        self.v_lin1 = nn.Linear(units, units)
        self.v_lin2 = nn.Linear(units, units)
        self.v_lin3 = nn.Linear(units, units)
        self.v_lin4 = nn.Linear(units, units)
        self.v_bn = gnn.BatchNorm(units)

        # Edge updates
        self.e_lin = nn.Linear(units, units)
        self.e_bn = gnn.BatchNorm(units)

    def forward(self, x, edge_index, edge_attr):
        x0 = x
        w0 = w = edge_attr

        # Vertex updates
        x1 = self.v_lin1(x0)
        x2 = self.v_lin2(x0)
        x3 = self.v_lin3(x0)
        x4 = self.v_lin4(x0)
        x = x0 + self.act_fn(
            self.v_bn(
                x1 + self.agg_fn(torch.sigmoid(w0) * x2[edge_index[1]], edge_index[0])
            )
        )

        # Edge updates
        w1 = self.e_lin(w0)
        w = w0 + self.act_fn(self.e_bn(w1 + x3[edge_index[0]] + x4[edge_index[1]]))
        return x, w


class GNNEncoder(nn.Module):
    """Anisotropic Graph Neural Network encoder with edge-gating mechanism as in Joshi et al. (2022)

    Args:
        num_layers: The number of GNN layers to stack in the network.
        embed_dim: The dimensionality of the embeddings for each node in the graph.
        act_fn: The activation function to use in each GNNLayer, see https://pytorch.org/docs/stable/nn.functional.html#non-linear-activation-functions for available options. Defaults to 'silu'.
        agg_fn: The aggregation function to use in each GNNLayer for pooling features. Options: 'add', 'mean', 'max'. Defaults to 'mean'.
    """

    def __init__(self, num_layers: int, embed_dim: int, act_fn="silu", agg_fn="mean"):
        super(GNNEncoder, self).__init__()
        self.act_fn = getattr(nn.functional, act_fn)
        self.agg_fn = agg_fn

        # Stack of GNN layers
        self.layers = nn.ModuleList(
            [GNNLayer(embed_dim, act_fn, agg_fn) for _ in range(num_layers)]
        )

    def forward(self, x, edge_index, w):
        """Sequentially passes the input graph data through the stacked GNN layers,
        applying specified transformations and aggregations to learn graph representations.

        Args:
            x: The node features of the graph with shape [num_nodes, embed_dim].
            edge_index: The edge indices of the graph with shape [2, num_edges].
            w: The edge attributes or weights with shape [num_edges, embed_dim].
        """
        x = self.act_fn(x)
        w = self.act_fn(w)
        for layer in self.layers:
            x, w = layer(x, edge_index, w)
        return x, w



import torch

from tensordict.tensordict import TensorDict

from rl4co.envs.routing.cvrp.env import CVRPEnv
from rl4co.utils.ops import gather_by_index
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class CVRPMVCEnv(CVRPEnv):
    """Capacitated Vehicle Routing Problem (CVRP) with maximum vehicle constraint environment."""

    name = "cvrpmvc"

    def _step(self, td: TensorDict) -> TensorDict:
        vehicles_used = td["vehicles_used"] + (
            (td["action"].unsqueeze(-1) == 0) & (td["current_node"] != 0)
        )

        current_node = td["action"][:, None]  # Add dimension for step
        n_loc = td["demand"].size(-1)  # Excludes depot

        # Not selected_demand is demand of first node (by clamp) so incorrect for nodes that visit depot!
        selected_demand = gather_by_index(
            td["demand"], torch.clamp(current_node - 1, 0, n_loc - 1), squeeze=False
        )

        # Increase capacity if depot is not visited, otherwise set to 0
        used_capacity = (td["used_capacity"] + selected_demand) * (
            current_node != 0
        ).float()

        demand_remaining = td["demand_remaining"] - selected_demand

        # Note: here we do not subtract one as we have to scatter so the first column allows scattering depot
        # Add one dimension since we write a single value
        visited = td["visited"].scatter(-1, current_node, 1)

        # SECTION: get done
        done = visited.sum(-1) == visited.size(-1)
        reward = torch.zeros_like(done)

        td.update(
            {
                "current_node": current_node,
                "used_capacity": used_capacity,
                "vehicles_used": vehicles_used,
                "demand_remaining": demand_remaining,
                "visited": visited,
                "reward": reward,
                "done": done,
            }
        )
        td.set("action_mask", self.get_action_mask(td))
        return td

    def _reset(
        self, td: TensorDict | None = None, batch_size: list | None = None
    ) -> TensorDict:
        td = super()._reset(td, batch_size)
        batch_size = batch_size or list(td.batch_size)
        td.set(
            "vehicles_used",
            torch.ones((*batch_size, 1), dtype=torch.int, device=td.device),
        )
        td.set("demand_remaining", td["demand"].sum(-1, keepdim=True))
        td.set(
            "max_vehicle", torch.ceil(td["demand_remaining"] / td["vehicle_capacity"]) + 1
        )
        return td

    @staticmethod
    def get_action_mask(td: TensorDict) -> torch.Tensor:
        # For demand steps_dim is inserted by indexing with id, for used_capacity insert node dim for broadcasting
        exceeds_cap = td["demand"] + td["used_capacity"] > td["vehicle_capacity"]

        # Nodes that cannot be visited are already visited or too much demand to be served now
        mask_loc = td["visited"][..., 1:].to(exceeds_cap.dtype) | exceeds_cap

        if "vehicles_used" in td.keys():
            max_vehicle = td["max_vehicle"]
            demand_remaining = td["demand_remaining"]
            capacity_remaining = (max_vehicle - td["vehicles_used"]) * td[
                "vehicle_capacity"
            ]
            mask_depot = (  # mask the depot
                (td["current_node"] == 0)  # if the depot is just visited
                | (
                    demand_remaining > capacity_remaining
                )  # or the unassigned vehicles' capacity can't sastify remaining demands
            ) & ~torch.all(
                mask_loc, dim=-1, keepdim=True
            )  # unless there's no other choices
        else:
            # Cannot visit the depot if just visited and still unserved nodes
            mask_depot = (td["current_node"] == 0) & ~torch.all(
                mask_loc, dim=-1, keepdim=True
            )
        return ~torch.cat((mask_depot, mask_loc), -1)



from typing import Callable

import torch

from tensordict.tensordict import TensorDict
from torch.distributions import Uniform

from rl4co.envs.common.utils import Generator, get_sampler
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class ATSPGenerator(Generator):
    """Data generator for the Asymmetric Travelling Salesman Problem (ATSP)
    Generate distance matrices inspired by the reference MatNet (Kwon et al., 2021)
    We satifsy the triangle inequality (TMAT class) in a batch

    Args:
        num_loc: number of locations (customers) in the TSP
        min_dist: minimum value for the distance between nodes
        max_dist: maximum value for the distance between nodes
        dist_distribution: distribution for the distance between nodes
        tmat_class: whether to generate a class of distance matrix

    Returns:
        A TensorDict with the following keys:
            locs [batch_size, num_loc, 2]: locations of each customer
    """

    def __init__(
        self,
        num_loc: int = 10,
        min_dist: float = 0.0,
        max_dist: float = 1.0,
        dist_distribution: int | float | str | type | Callable = Uniform,
        tmat_class: bool = True,
        **kwargs,
    ):
        self.num_loc = num_loc
        self.min_dist = min_dist
        self.max_dist = max_dist
        self.tmat_class = tmat_class

        # Distance distribution
        if kwargs.get("dist_sampler", None) is not None:
            self.dist_sampler = kwargs["dist_sampler"]
        else:
            self.dist_sampler = get_sampler("dist", dist_distribution, 0.0, 1.0, **kwargs)

    def _generate(self, batch_size) -> TensorDict:
        # Generate distance matrices inspired by the reference MatNet (Kwon et al., 2021)
        # We satifsy the triangle inequality (TMAT class) in a batch
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size
        dms = (
            self.dist_sampler.sample((batch_size + [self.num_loc, self.num_loc]))
            * (self.max_dist - self.min_dist)
            + self.min_dist
        )
        dms[..., torch.arange(self.num_loc), torch.arange(self.num_loc)] = 0
        log.info("Using TMAT class (triangle inequality): {}".format(self.tmat_class))
        if self.tmat_class:
            for i in range(self.num_loc):
                dms = torch.minimum(dms, dms[..., :, [i]] + dms[..., [i], :])
        return TensorDict({"cost_matrix": dms}, batch_size=batch_size)



import matplotlib.pyplot as plt
import numpy as np
import torch

from matplotlib import cm, colormaps

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(td, actions=None, ax=None):
    # TODO: color switch with new agents; add pickup and delivery nodes as in `PDPEnv.render`

    num_routine = (actions == 0).sum().item() + 2
    base = colormaps["nipy_spectral"]
    color_list = base(np.linspace(0, 1, num_routine))
    cmap_name = base.name + str(num_routine)
    out = base.from_list(cmap_name, color_list, num_routine)

    if ax is None:
        # Create a plot of the nodes
        _, ax = plt.subplots()

    td = td.detach().cpu()

    if actions is None:
        actions = td.get("action", None)

    # if batch_size greater than 0 , we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]
        actions = actions[0]

    locs = td["locs"]

    # add the depot at the first action and the end action
    actions = torch.cat([torch.tensor([0]), actions, torch.tensor([0])])

    # gather locs in order of action if available
    if actions is None:
        log.warning("No action in TensorDict, rendering unsorted locs")
    else:
        locs = locs

    # Cat the first node to the end to complete the tour
    x, y = locs[:, 0], locs[:, 1]

    # plot depot
    ax.scatter(
        locs[0, 0],
        locs[0, 1],
        edgecolors=cm.Set2(2),
        facecolors="none",
        s=100,
        linewidths=2,
        marker="s",
        alpha=1,
    )

    # plot visited nodes
    ax.scatter(
        x[1:],
        y[1:],
        edgecolors=cm.Set2(0),
        facecolors="none",
        s=50,
        linewidths=2,
        marker="o",
        alpha=1,
    )

    # text depot
    ax.text(
        locs[0, 0],
        locs[0, 1] - 0.025,
        "Depot",
        horizontalalignment="center",
        verticalalignment="top",
        fontsize=10,
        color=cm.Set2(2),
    )

    # plot actions
    color_idx = 0
    for action_idx in range(len(actions) - 1):
        if actions[action_idx] == 0:
            color_idx += 1
        from_loc = locs[actions[action_idx]]
        to_loc = locs[actions[action_idx + 1]]
        ax.plot(
            [from_loc[0], to_loc[0]],
            [from_loc[1], to_loc[1]],
            color=out(color_idx),
            lw=1,
        )
        ax.annotate(
            "",
            xy=(to_loc[0], to_loc[1]),
            xytext=(from_loc[0], from_loc[1]),
            arrowprops=dict(arrowstyle="-|>", color=out(color_idx)),
            size=15,
            annotation_clip=False,
        )

    return ax



from rl4co.models.common.constructive.autoregressive import (
    AutoregressiveDecoder,
    AutoregressiveEncoder,
    AutoregressivePolicy,
)
from rl4co.models.common.constructive.base import (
    ConstructiveDecoder,
    ConstructiveEncoder,
    ConstructivePolicy,
)
from rl4co.models.common.constructive.nonautoregressive import (
    NonAutoregressiveDecoder,
    NonAutoregressiveEncoder,
    NonAutoregressivePolicy,
)
from rl4co.models.common.improvement import (
    ImprovementDecoder,
    ImprovementEncoder,
    ImprovementPolicy,
)
from rl4co.models.common.transductive.base import TransductiveModel



from rl4co.models.rl.a2c.a2c import A2C
from rl4co.models.rl.common.base import RL4COLitModule
from rl4co.models.rl.ppo.n_step_ppo import n_step_PPO
from rl4co.models.rl.ppo.ppo import PPO
from rl4co.models.rl.ppo.stepwise_ppo import StepwisePPO
from rl4co.models.rl.reinforce.reinforce import REINFORCE



from typing import Any, Generator, NamedTuple, Optional, Union

import torch

from tensordict import TensorDict

from rl4co.utils.ops import gather_by_index


class SHPPMapping(NamedTuple):
    map_action_index: torch.Tensor
    map_node_index: torch.Tensor
    subprob_coordinates: torch.Tensor


class TSP2SHPPAdapter:
    """Adapter class for decomposing and composing TSP solutions."""

    subproblem_env_name = "shpp"

    def __init__(
        self,
        parent_td: TensorDict,
        actions: torch.Tensor,
        subprob_batch_size: Optional[int] = None,
        partition_node_count: Union[int, list[int]] = 20,
        shift: int = 0,
    ) -> None:
        batch_size = parent_td.batch_size[0]
        n_samples = actions.shape[0] // batch_size
        assert actions.shape[0] == n_samples * batch_size

        self._actions = actions.cpu().clone()
        self.shpp_node_counts = (
            [partition_node_count]
            if isinstance(partition_node_count, int)
            else partition_node_count
        )
        self.coordinates = parent_td["locs"].cpu()
        self.subprob_batch_size = subprob_batch_size
        self.shift = shift

    def _get_batched_subprobs_one_iter(
        self, node_count: int
    ) -> Generator[SHPPMapping, Any, None]:
        self.shpp_actions, shpp_coordinates, self.share_memory = self.action_partitioner(
            self._actions, self.coordinates, node_count
        )
        self.shpp_node_count = node_count
        shpp_count = len(self.shpp_actions)
        if shpp_count == 0:
            return
        batch_size = self.subprob_batch_size is None or shpp_count
        for start_index in range(0, shpp_count, batch_size):
            map_action_index = torch.arange(
                start_index, min(start_index + batch_size, shpp_count)
            )
            map_node_index = self.shpp_actions[map_action_index]  # shpp_index
            this_shpp_coordinates = shpp_coordinates[map_action_index]
            yield SHPPMapping(map_action_index, map_node_index, this_shpp_coordinates)

    def get_batched_subprobs(self) -> Generator[SHPPMapping, Any, None]:
        if len(self.shpp_node_counts) == 1:
            yield from self._get_batched_subprobs_one_iter(self.shpp_node_counts[0])
        else:
            for node_count in self.shpp_node_counts:
                yield from self._get_batched_subprobs_one_iter(node_count)
                self.get_actions()  # update self._actions in case not shared
                if self.shift:
                    self._actions = torch.roll(self._actions, self.shift, 1)

    def update_actions(self, mapping: SHPPMapping, subtsp_actions: torch.Tensor):
        self.shpp_actions[mapping.map_action_index] = gather_by_index(
            mapping.map_node_index, subtsp_actions
        )

    def get_actions(self):
        if (
            not self.share_memory and len(self.shpp_actions) > 0
        ):  # write back to actions if memory is not shared
            tsp_nodes = self._actions.shape[1]
            tsp_nodes -= tsp_nodes % self.shpp_node_count
            self._actions[:, :tsp_nodes] = self.shpp_actions.view(-1, tsp_nodes)
        return self._actions

    @staticmethod
    def action_partitioner(
        actions: torch.Tensor, coordinates: torch.Tensor, shpp_nodes: int
    ):
        batch_size, tsp_nodes, _ = coordinates.shape
        share_memory = tsp_nodes % shpp_nodes == 0
        if share_memory:
            shpp_actions = actions.view(-1, shpp_nodes)
        else:
            tsp_nodes -= tsp_nodes % shpp_nodes
            shpp_actions = actions[:, :tsp_nodes].reshape(
                -1, shpp_nodes
            )  # trim tail nodes

        repeated_coordinates = (
            coordinates.unsqueeze(1)
            .expand(
                coordinates.shape[0],
                len(shpp_actions) // batch_size,
                *coordinates.shape[1:],
            )
            .flatten(0, 1)
        )  # (bs, n_nodes, 2) -> (bs*n_samples*shpp_per_route, n_nodes, 2)
        shpp_coordinates = gather_by_index(repeated_coordinates, shpp_actions)

        return shpp_actions, shpp_coordinates, share_memory



import torch.nn as nn

from rl4co.envs import RL4COEnvBase
from rl4co.models.common.constructive.autoregressive.policy import AutoregressivePolicy
from rl4co.models.zoo.am.encoder import AttentionModelEncoder
from rl4co.models.zoo.matnet.encoder import MatNetEncoder
from rl4co.models.zoo.polynet.decoder import PolyNetDecoder


class PolyNetPolicy(AutoregressivePolicy):
    """
    # TODO
    Polynet policy based on Hottung et al. (2024) https://arxiv.org/abs/2402.14048.
    The model uses either the AttentionModel encoder or the MatNet encoder in combination with
    a custom PolyNet decoder.

    Note: The default arguments for the AttentionModel encoder follow the POMO paper. The default decoding type
    during validation and testing is 'sampling'.

    Args:
        k: Number of strategies to learn ("K" in the paper)
        encoder_type: Type of encoder that should be used. "AM" or "MatNet" are supported.
        embed_dim: Dimension of the node embeddings
        num_encoder_layers: Number of layers in the encoder
        num_heads: Number of heads in the attention layers
        normalization: Normalization type in the attention layers
        feedforward_hidden: Dimension of the hidden layer in the feedforward network
        env_name: Name of the environment used to initialize embeddings
        temperature: Temperature for the softmax
        tanh_clipping: Tanh clipping value (see Bello et al., 2016)
        mask_logits: Whether to mask the logits during decoding
        train_decode_type: Type of decoding to use during training
        val_decode_type: Type of decoding to use during validation
        test_decode_type: Type of decoding to use during testing
        **kwargs: keyword arguments passed to the encoder and decoder modules
    """

    def __init__(
        self,
        k: int,
        encoder: nn.Module = None,
        encoder_type: str = "AM",
        embed_dim: int = 128,
        num_encoder_layers: int = 6,
        num_heads: int = 8,
        normalization: str = "instance",
        feedforward_hidden: int = 512,
        env_name: str | RL4COEnvBase = "tsp",
        temperature: float = 1.0,
        tanh_clipping: float = 10.0,
        mask_logits: bool = True,
        train_decode_type: str = "sampling",
        val_decode_type: str = "sampling",
        test_decode_type: str = "sampling",
        **kwargs,
    ):
        if encoder is None:
            if encoder_type == "AM":
                encoder = AttentionModelEncoder(
                    embed_dim=embed_dim,
                    num_heads=num_heads,
                    num_layers=num_encoder_layers,
                    env_name=env_name,
                    normalization=normalization,
                    feedforward_hidden=feedforward_hidden,
                    **kwargs,
                )
            elif encoder_type == "MatNet":
                kwargs_with_defaults = {"init_embedding_kwargs": {"mode": "RandomOneHot"}}
                kwargs_with_defaults.update(kwargs)
                encoder = MatNetEncoder(
                    embed_dim=embed_dim,
                    num_heads=num_heads,
                    num_layers=num_encoder_layers,
                    normalization=normalization,
                    **kwargs_with_defaults,
                )

        decoder = PolyNetDecoder(
            k=k,
            encoder_type=encoder_type,
            embed_dim=embed_dim,
            num_heads=num_heads,
            env_name=env_name,
            **kwargs,
        )

        super(PolyNetPolicy, self).__init__(
            encoder=encoder,
            decoder=decoder,
            env_name=env_name,
            temperature=temperature,
            tanh_clipping=tanh_clipping,
            mask_logits=mask_logits,
            train_decode_type=train_decode_type,
            val_decode_type=val_decode_type,
            test_decode_type=test_decode_type,
            **kwargs,
        )



import torch

from tensordict.tensordict import TensorDict


def render(td: TensorDict, actions=None, ax=None):
    import matplotlib.pyplot as plt

    markersize = 8

    td = td.detach().cpu()

    # If batch_size greater than 0, we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]
        if actions is not None:
            actions = actions[0]

    n_depots = td["capacity"].shape[-1]
    n_pickups = (td["locs"].size(-2) - n_depots) // 2

    # Variables
    init_deliveries = td["to_deliver"][n_depots:]
    delivery_locs = td["locs"][n_depots:][~init_deliveries.bool()]
    pickup_locs = td["locs"][n_depots:][init_deliveries.bool()]
    depot_locs = td["locs"][:n_depots]
    actions = actions if actions is not None else td["action"]

    if ax is None:
        _, ax = plt.subplots(figsize=(4, 4))

    # Split actions into tours (segments starting from depot)
    tours = []
    current_tour = []
    current_depot = actions[0]  # Start from first depot

    for action in actions:
        if action < n_depots:
            if current_tour:  # If we have a non-empty tour
                # Add the current tour with its starting depot
                tours.append((current_depot, current_tour))
            current_depot = action
            current_tour = []
        else:
            current_tour.append(action)

    # Add the last tour if it exists
    if current_tour:
        tours.append((current_depot, current_tour))

    # Plot each tour with a different color
    for tour_idx, (depot, tour) in enumerate(tours):
        color = f"C{tour_idx}"  # Use matplotlib's color cycle

        # Plot from depot to first location
        if tour:  # Only if tour is not empty
            from_loc = td["locs"][depot]
            to_loc = td["locs"][tour[0]]
            ax.plot([from_loc[0], to_loc[0]], [from_loc[1], to_loc[1]], color=color)
            ax.annotate(
                "",
                xy=(to_loc[0], to_loc[1]),
                xytext=(from_loc[0], from_loc[1]),
                arrowprops=dict(arrowstyle="->", color=color),
                annotation_clip=False,
            )

            # Plot connections between tour locations
            for i in range(len(tour) - 1):
                from_loc = td["locs"][tour[i]]
                to_loc = td["locs"][tour[i + 1]]
                ax.plot([from_loc[0], to_loc[0]], [from_loc[1], to_loc[1]], color=color)
                ax.annotate(
                    "",
                    xy=(to_loc[0], to_loc[1]),
                    xytext=(from_loc[0], from_loc[1]),
                    arrowprops=dict(arrowstyle="->", color=color),
                    annotation_clip=False,
                )

            # Plot return to depot in faint grey dashed line
            from_loc = td["locs"][tour[-1]]
            to_loc = td["locs"][depot]
            ax.plot(
                [from_loc[0], to_loc[0]],
                [from_loc[1], to_loc[1]],
                color="grey",
                linestyle="--",
                alpha=0.3,
            )
            ax.annotate(
                "",
                xy=(to_loc[0], to_loc[1]),
                xytext=(from_loc[0], from_loc[1]),
                arrowprops=dict(arrowstyle="->", color="grey", alpha=0.3),
                annotation_clip=False,
            )

    # Annotate node locations
    for i, loc in enumerate(td["locs"]):
        ax.annotate(
            str(i),
            (loc[0], loc[1]),
            textcoords="offset points",
            xytext=(0, 5),
            ha="center",
        )

    # Plot depots
    for i, depot_loc in enumerate(depot_locs):
        ax.plot(
            depot_loc[0],
            depot_loc[1],
            "tab:green",
            marker="s",
            markersize=markersize,
            label="Depot" if i == 0 else None,
        )

    # Plot pickup locations
    for i, pickup_loc in enumerate(pickup_locs):
        ax.plot(
            pickup_loc[0],
            pickup_loc[1],
            "tab:red",
            marker="^",
            markersize=markersize,
            label="Pickup" if i == 0 else None,
        )

    # Plot delivery locations
    for i, delivery_loc in enumerate(delivery_locs):
        ax.plot(
            delivery_loc[0],
            delivery_loc[1],
            "tab:blue",
            marker="x",
            markersize=markersize,
            label="Delivery" if i == 0 else None,
        )

    # Plot pickup and delivery pairs
    for i in range(n_pickups):
        pickup_loc = td["locs"][n_depots + i]
        delivery_loc = td["locs"][n_depots + n_pickups + i]
        ax.plot(
            [pickup_loc[0], delivery_loc[0]],
            [pickup_loc[1], delivery_loc[1]],
            "k--",
            alpha=0.5,
        )

    return ax



import torch
import torch.nn as nn


class EASLayerNet(nn.Module):
    """Instantiate weights and biases for the added layer.
    The layer is defined as: h = relu(emb * W1 + b1); out = h * W2 + b2.
    Wrapping in `nn.Parameter` makes the parameters trainable and sets gradient to True.

    Args:
        num_instances: Number of instances in the dataset
        emb_dim: Dimension of the embedding
    """

    def __init__(self, num_instances: int, emb_dim: int):
        super().__init__()
        # W2 and b2 are initialized to zero so in the first iteration the layer is identity
        self.W1 = nn.Parameter(torch.randn(num_instances, emb_dim, emb_dim))
        self.b1 = nn.Parameter(torch.randn(num_instances, 1, emb_dim))
        self.W2 = nn.Parameter(torch.zeros(num_instances, emb_dim, emb_dim))
        self.b2 = nn.Parameter(torch.zeros(num_instances, 1, emb_dim))
        torch.nn.init.xavier_uniform_(self.W1)
        torch.nn.init.xavier_uniform_(self.b1)

    def forward(self, *args):
        """emb: [num_instances, group_num, emb_dim]"""
        # get tensor arg (from partial instantiation)
        emb = [arg for arg in args if isinstance(arg, torch.Tensor)][0]
        h = torch.relu(torch.matmul(emb, self.W1) + self.b1.expand_as(emb))
        return torch.matmul(h, self.W2) + self.b2.expand_as(h)



import abc

from os.path import join as pjoin
from typing import Iterable, Optional

import torch

from tensordict.tensordict import TensorDict
from torchrl.envs import EnvBase

from rl4co.data.dataset import TensorDictDataset
from rl4co.data.utils import load_npz_to_tensordict
from rl4co.utils.ops import get_num_starts, select_start_nodes
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class RL4COEnvBase(EnvBase, metaclass=abc.ABCMeta):
    """Base class for RL4CO environments based on TorchRL EnvBase.
    The environment has the usual methods for stepping, resetting, and getting the specifications of the environment
    that shoud be implemented by the subclasses of this class.
    It also has methods for getting the reward, action mask, and checking the validity of the solution, and
    for generating and loading the datasets (supporting multiple dataloaders as well for validation and testing).

    Args:
        data_dir: Root directory for the dataset
        train_file: Name of the training file
        val_file: Name of the validation file
        test_file: Name of the test file
        val_dataloader_names: Names of the dataloaders to use for validation
        test_dataloader_names: Names of the dataloaders to use for testing
        check_solution: Whether to check the validity of the solution at the end of the episode
        dataset_cls: Dataset class to use for the environment (which can influence performance)
        seed: Seed for the environment
        device: Device to use. Generally, no need to set as tensors are updated on the fly
        batch_size: Batch size to use for the environment. Generally, no need to set as tensors are updated on the fly
        run_type_checks: If True, run type checks on the TensorDicts at each step
        allow_done_after_reset: If True, an environment can be done after a reset
        _torchrl_mode: Whether to use the TorchRL mode (see :meth:`step` for more details)
    """

    batch_locked = False

    def __init__(
        self,
        *,
        data_dir: str = "data/",
        train_file: str = None,
        val_file: str = None,
        test_file: str = None,
        val_dataloader_names: list = None,
        test_dataloader_names: list = None,
        check_solution: bool = True,
        dataset_cls: callable = TensorDictDataset,
        seed: int = None,
        device: str = "cpu",
        batch_size: torch.Size = None,
        run_type_checks: bool = False,
        allow_done_after_reset: bool = False,
        _torchrl_mode: bool = False,
        **kwargs,
    ):
        super().__init__(
            device=device,
            batch_size=batch_size,
            run_type_checks=run_type_checks,
            allow_done_after_reset=allow_done_after_reset,
        )
        # if any kwargs are left, we want to warn the user
        kwargs.pop("name", None)  # we remove the name for checking
        if kwargs:
            log.error(
                f"Unused keyword arguments: {', '.join(kwargs.keys())}. "
                "Please check the base class documentation at https://rl4co.readthedocs.io/en/latest/_content/api/envs/base.html. "
                "In case you would like to pass data generation arguments, please pass a `generator` method instead "
                "or for example: `generator_kwargs=dict(num_loc=50)` to the constructor."
            )
        self.data_dir = data_dir
        self.train_file = pjoin(data_dir, train_file) if train_file is not None else None
        self._torchrl_mode = _torchrl_mode
        self.dataset_cls = dataset_cls

        def get_files(f):
            if f is not None:
                if isinstance(f, Iterable) and not isinstance(f, str):
                    return [pjoin(data_dir, _f) for _f in f]
                else:
                    return pjoin(data_dir, f)
            return None

        def get_multiple_dataloader_names(f, names):
            if f is not None:
                if isinstance(f, Iterable) and not isinstance(f, str):
                    if names is None:
                        names = [f"{i}" for i in range(len(f))]
                    else:
                        assert len(names) == len(
                            f
                        ), "Number of dataloader names must match number of files"
                else:
                    if names is not None:
                        log.warning(
                            "Ignoring dataloader names since only one dataloader is provided"
                        )
            return names

        self.val_file = get_files(val_file)
        self.test_file = get_files(test_file)
        self.val_dataloader_names = get_multiple_dataloader_names(
            self.val_file, val_dataloader_names
        )
        self.test_dataloader_names = get_multiple_dataloader_names(
            self.test_file, test_dataloader_names
        )
        self.check_solution = check_solution
        if seed is None:
            seed = torch.empty((), dtype=torch.int64).random_().item()
        self.set_seed(seed)

    def step(self, td: TensorDict) -> TensorDict:
        """Step function to call at each step of the episode containing an action.
        If `_torchrl_mode` is True, we call `_torchrl_step` instead which set the
        `next` key of the TensorDict to the next state - this is the usual way to do it in TorchRL,
        but inefficient in our case
        """
        if not self._torchrl_mode:
            # Default: just return the TensorDict without farther checks etc is faster
            td = self._step(td)
            return {"next": td}
        else:
            # Since we simplify the syntax
            return self._torchrl_step(td)

    def reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        """Reset function to call at the beginning of each episode"""
        if batch_size is None:
            batch_size = self.batch_size if td is None else td.batch_size
        if td is None or td.is_empty():
            td = self.generator(batch_size=batch_size)
        batch_size = [batch_size] if isinstance(batch_size, int) else batch_size
        self.to(td.device)
        return super().reset(td, batch_size=batch_size)

    def _torchrl_step(self, td: TensorDict) -> TensorDict:
        """See :meth:`super().step` for more details.
        This is the usual way to do it in TorchRL, but inefficient in our case

        Note:
            Here we clone the TensorDict to avoid recursion error, since we allow
            for directly updating the TensorDict in the step function
        """
        # sanity check
        self._assert_tensordict_shape(td)
        next_preset = td.get("next", None)

        next_tensordict = self._step(
            td.clone()
        )  # NOTE: we clone to avoid recursion error
        next_tensordict = self._step_proc_data(next_tensordict)
        if next_preset is not None:
            next_tensordict.update(next_preset.exclude(*next_tensordict.keys(True, True)))
        td.set("next", next_tensordict)
        return td

    @abc.abstractmethod
    def _step(self, td: TensorDict) -> TensorDict:
        """Step function to call at each step of the episode containing an action.
        Gives the next observation, reward, done
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _reset(self, td: Optional[TensorDict] = None, batch_size=None) -> TensorDict:
        """Reset function to call at the beginning of each episode"""
        raise NotImplementedError

    def _make_spec(self, td_params: TensorDict = None):
        """Make the specifications of the environment (observation, action, reward, done)"""
        raise NotImplementedError

    def get_reward(
        self, td: TensorDict, actions: torch.Tensor, check_solution: Optional[bool] = None
    ) -> torch.Tensor:
        """Function to compute the reward. Can be called by the agent to compute the reward of the current state
        This is faster than calling step() and getting the reward from the returned TensorDict at each time for CO tasks
        """
        # Fallback to env setting if not assigned
        check_solution = self.check_solution if check_solution is None else check_solution
        if check_solution:
            self.check_solution_validity(td, actions)
        return self._get_reward(td, actions)

    @abc.abstractmethod
    def _get_reward(self, td, actions) -> TensorDict:
        """Function to compute the reward. Can be called by the agent to compute the reward of the current state
        This is faster than calling step() and getting the reward from the returned TensorDict at each time for CO tasks
        """
        raise NotImplementedError

    def get_action_mask(self, td: TensorDict) -> torch.Tensor:
        """Function to compute the action mask (feasible actions) for the current state
        Action mask is 1 if the action is feasible, 0 otherwise
        """
        raise NotImplementedError

    def get_num_starts(self, td):
        return get_num_starts(td, self.name)

    def select_start_nodes(self, td, num_starts):
        return select_start_nodes(td, self, num_starts)

    def check_solution_validity(self, td: TensorDict, actions: torch.Tensor) -> None:
        """Function to check whether the solution is valid. Can be called by the agent to check the validity of the current state
        This is called with the full solution (i.e. all actions) at the end of the episode
        """
        raise NotImplementedError

    def replace_selected_actions(
        self,
        cur_actions: torch.Tensor,
        new_actions: torch.Tensor,
        selection_mask: torch.Tensor,
    ) -> torch.Tensor:
        """
        Replace selected current actions with updated actions based on `selection_mask`.
        """
        raise NotImplementedError

    def local_search(
        self, td: TensorDict, actions: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """Function to improve the solution. Can be called by the agent to improve the current state
        This is called with the full solution (i.e. all actions) at the end of the episode
        """
        raise NotImplementedError(
            f"Local is not implemented yet for {self.name} environment"
        )

    def dataset(self, batch_size=[], phase="train", filename=None):
        """Return a dataset of observations
        Generates the dataset if it does not exist, otherwise loads it from file
        """
        if filename is not None:
            log.info(f"Overriding dataset filename from {filename}")
        f = getattr(self, f"{phase}_file") if filename is None else filename
        if f is None:
            if phase != "train":
                log.warning(f"{phase}_file not set. Generating dataset instead")
            td = self.generator(batch_size)
        else:
            log.info(f"Loading {phase} dataset from {f}")
            if phase == "train":
                log.warning(
                    "Loading training dataset from file. This may not be desired in RL since "
                    "the dataset is fixed and the agent will not be able to explore new states"
                )
            try:
                if isinstance(f, Iterable) and not isinstance(f, str):
                    names = getattr(self, f"{phase}_dataloader_names")
                    return {
                        name: self.dataset_cls(self.load_data(_f, batch_size))
                        for name, _f in zip(names, f)
                    }
                else:
                    td = self.load_data(f, batch_size)
            except FileNotFoundError:
                log.error(
                    f"Provided file name {f} not found. Make sure to provide a file in the right path first or "
                    f"unset {phase}_file to generate data automatically instead"
                )
                td = self.generator(batch_size)

        return self.dataset_cls(td)

    def transform(self):
        """Used for converting TensorDict variables (such as with torch.cat) efficiently
        https://pytorch.org/rl/reference/generated/torchrl.envs.transforms.Transform.html
        By default, we do not need to transform the environment since we use specific embeddings
        """
        return self

    def render(self, *args, **kwargs):
        """Render the environment"""
        raise NotImplementedError(
            f"Render is not implemented for {self.name} environment. Please implement the `render` method in the subclass."
        )

    @staticmethod
    def load_data(fpath, batch_size=[]):
        """Dataset loading from file"""
        return load_npz_to_tensordict(fpath)

    def _set_seed(self, seed: Optional[int]):
        """Set the seed for the environment"""
        rng = torch.manual_seed(seed)
        self.rng = rng

    def to(self, device):
        """Override `to` device method for safety against `None` device (may be found in `TensorDict`)"""
        if device is None:
            return self
        else:
            return super().to(device)

    @staticmethod
    def solve(
        instances: TensorDict,
        max_runtime: float,
        num_procs: int = 1,
        **kwargs,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Classical solver for the environment. This is a wrapper for the baselines solver.

        Args:
            instances: The instances to solve
            max_runtime: The maximum runtime for the solver
            num_procs: The number of processes to use

        Returns:
            A tuple containing the action and the cost, respectively
        """
        raise NotImplementedError

    def __getstate__(self):
        """Return the state of the environment. By default, we want to avoid pickling
        the random number generator directly as it is not allowed by `deepcopy`
        """
        state = self.__dict__.copy()
        state["rng"] = state["rng"].get_state()
        return state

    def __setstate__(self, state):
        """Set the state of the environment. By default, we want to avoid pickling
        the random number generator directly as it is not allowed by `deepcopy`
        """
        self.__dict__.update(state)
        self.rng = torch.manual_seed(0)
        self.rng.set_state(state["rng"])


class ImprovementEnvBase(RL4COEnvBase, metaclass=abc.ABCMeta):
    """Base class for Improvement environments based on RL4CO EnvBase.
    Note that this class assumes that the solution is stored in a linked list format.
    Here, if `rec[i] = j`, it means the node `i` is connected to node `j`, i.e., edge `i-j` is in the solution.
    For example, if edge `0-1`, edge `1-5`, edge `2-10` are in the solution, so we have `rec[0]=1`, `rec[1]=5` and `rec[2]=10`.
    Kindly see https://github.com/yining043/VRP-DACT/blob/new_version/Play_with_DACT.ipynb for an example at the end for TSP.
    """

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__(**kwargs)

    @abc.abstractmethod
    def _step(self, td: TensorDict, solution_to=None) -> TensorDict:
        raise NotImplementedError

    def step_to_solution(self, td, solution) -> TensorDict:
        return self._step(td, solution_to=solution)

    @staticmethod
    def _get_reward(td, actions) -> TensorDict:
        raise NotImplementedError(
            "This function is not used for improvement tasks since the reward is computed per step"
        )

    @staticmethod
    def get_costs(coordinates, rec):
        batch_size, size = rec.size()

        # calculate the route length value
        d1 = coordinates.gather(1, rec.long().unsqueeze(-1).expand(batch_size, size, 2))
        d2 = coordinates
        length = (d1 - d2).norm(p=2, dim=2).sum(1)

        return length

    @staticmethod
    def _get_real_solution(rec):
        batch_size, seq_length = rec.size()
        visited_time = torch.zeros((batch_size, seq_length)).to(rec.device)
        pre = torch.zeros((batch_size), device=rec.device).long()
        for i in range(seq_length):
            visited_time[torch.arange(batch_size), rec[torch.arange(batch_size), pre]] = (
                i + 1
            )
            pre = rec[torch.arange(batch_size), pre]

        visited_time = visited_time % seq_length
        return visited_time.argsort()

    @staticmethod
    def _get_linked_list_solution(solution):
        solution_pre = solution
        solution_post = torch.cat((solution[:, 1:], solution[:, :1]), 1)

        rec = solution.clone()
        rec.scatter_(1, solution_pre, solution_post)
        return rec

    @classmethod
    def get_best_solution(cls, td):
        return cls._get_real_solution(td["rec_best"])

    @classmethod
    def get_current_solution(cls, td):
        return cls._get_real_solution(td["rec_current"])



import numpy as np
import pyvrp as pyvrp

from pyvrp import Client, Depot, ProblemData, VehicleType, solve as _solve
from pyvrp.constants import MAX_VALUE
from pyvrp.stop import MaxRuntime
from tensordict.tensordict import TensorDict
from torch import Tensor

from .constants import PYVRP_SCALING_FACTOR
from .utils import scale


def solve(instance: TensorDict, max_runtime: float, **kwargs) -> tuple[Tensor, Tensor]:
    """
    Solves the AnyVRP instance with PyVRP.

    Args:
        instance: The AnyVRP instance to solve.
        max_runtime: The maximum runtime for the solver.

    Returns:
        A tuple containing the action and the cost, respectively.
    """
    data = instance2data(instance, PYVRP_SCALING_FACTOR)
    stop = MaxRuntime(max_runtime)
    result = _solve(data, stop)

    solution = result.best
    action = solution2action(solution)
    cost = -result.cost() / PYVRP_SCALING_FACTOR

    return action, cost


def instance2data(instance: TensorDict, scaling_factor: int) -> ProblemData:
    """
    Converts an AnyVRP instance to a ProblemData instance.

    Args:
        instance: The AnyVRP instance to convert.
        scaling_factor: The scaling factor to use for the conversion.

    Returns:
        The ProblemData instance.
    """
    num_locs = instance["demand_backhaul"].size()[0]

    time_windows = scale(instance["time_windows"], scaling_factor)
    pickup = scale(instance["demand_backhaul"], scaling_factor)
    delivery = scale(instance["demand_linehaul"], scaling_factor)
    service = scale(instance["service_time"], scaling_factor)
    coords = scale(instance["locs"], scaling_factor)
    capacity = scale(instance["vehicle_capacity"], scaling_factor)
    max_distance = scale(instance["distance_limit"], scaling_factor)

    depot = Depot(
        x=coords[0][0],
        y=coords[0][1],
    )

    clients = [
        Client(
            x=coords[idx][0],
            y=coords[idx][1],
            tw_early=time_windows[idx][0],
            tw_late=time_windows[idx][1],
            delivery=delivery[idx],
            pickup=pickup[idx],
            service_duration=service[idx],
        )
        for idx in range(1, num_locs)
    ]

    vehicle_type = VehicleType(
        num_available=num_locs - 1,  # one vehicle per client
        capacity=capacity,
        max_distance=max_distance,
        tw_early=time_windows[0][0],
        tw_late=time_windows[0][1],
    )

    matrix = scale(instance["cost_matrix"], scaling_factor)

    if instance["open_route"]:
        # Vehicles do not need to return to the depot, so we set all arcs
        # to the depot to zero.
        matrix[:, 0] = 0

    if instance["backhaul_class"] == 1:  # VRP with backhauls
        # In VRPB, linehauls must be served before backhauls. This can be
        # enforced by setting a high value for the distance/duration from depot
        # to backhaul (forcing linehaul to be served first) and a large value
        # from backhaul to linehaul (avoiding linehaul after backhaul clients).
        linehaul = np.flatnonzero(delivery > 0)
        backhaul = np.flatnonzero(pickup > 0)
        # Note: we remove the constraint that we cannot visit backhauls *only* in a
        # a single route as per Slack discussion
        # matrix[0, backhaul] = MAX_VALUE
        matrix[np.ix_(backhaul, linehaul)] = MAX_VALUE

    return ProblemData(clients, [depot], [vehicle_type], [matrix], [matrix])


def solution2action(solution: pyvrp.Solution) -> list[int]:
    """
    Converts a PyVRP solution to the action representation, i.e., a giant tour.
    """
    return [visit for route in solution.routes() for visit in route.visits() + [0]]



from math import factorial

import torch
import torch.nn as nn

from tensordict import TensorDict

from rl4co.envs.scheduling.ffsp.env import FFSPEnv
from rl4co.models.common.constructive.autoregressive import AutoregressivePolicy
from rl4co.models.zoo.matnet.decoder import (
    MatNetDecoder,
    MatNetFFSPDecoder,
    MultiStageFFSPDecoder,
)
from rl4co.models.zoo.matnet.encoder import MatNetEncoder
from rl4co.utils.ops import batchify
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class MatNetPolicy(AutoregressivePolicy):
    """MatNet Policy from Kwon et al., 2021.
    Reference: https://arxiv.org/abs/2106.11113

    Warning:
        This implementation is under development and subject to change.

    Args:
        env_name: Name of the environment used to initialize embeddings
        embed_dim: Dimension of the node embeddings
        num_encoder_layers: Number of layers in the encoder
        num_heads: Number of heads in the attention layers
        normalization: Normalization type in the attention layers
        **kwargs: keyword arguments passed to the `AutoregressivePolicy`

    Default paarameters are adopted from the original implementation.
    """

    def __init__(
        self,
        env_name: str = "atsp",
        embed_dim: int = 256,
        num_encoder_layers: int = 5,
        num_heads: int = 16,
        normalization: str = "instance",
        init_embedding_kwargs: dict = {"mode": "RandomOneHot"},
        use_graph_context: bool = False,
        bias: bool = False,
        **kwargs,
    ):
        if env_name not in ["atsp", "ffsp"]:
            log.error(f"env_name {env_name} is not originally implemented in MatNet")

        if env_name == "ffsp":
            decoder = MatNetFFSPDecoder(
                embed_dim=embed_dim,
                num_heads=num_heads,
                use_graph_context=use_graph_context,
                out_bias=True,
            )

        else:
            decoder = MatNetDecoder(
                env_name=env_name,
                embed_dim=embed_dim,
                num_heads=num_heads,
                use_graph_context=use_graph_context,
            )

        super(MatNetPolicy, self).__init__(
            env_name=env_name,
            encoder=MatNetEncoder(
                embed_dim=embed_dim,
                num_heads=num_heads,
                num_layers=num_encoder_layers,
                normalization=normalization,
                init_embedding_kwargs=init_embedding_kwargs,
                bias=bias,
            ),
            decoder=decoder,
            embed_dim=embed_dim,
            num_encoder_layers=num_encoder_layers,
            num_heads=num_heads,
            normalization=normalization,
            **kwargs,
        )


class MultiStageFFSPPolicy(nn.Module):
    """Policy for solving the FFSP using a seperate encoder and decoder for each
    stage. This requires the 'while not td["done"].all()'-loop to be on policy level
    (instead of decoder level)."""

    def __init__(
        self,
        stage_cnt: int,
        embed_dim: int = 512,
        num_heads: int = 16,
        num_encoder_layers: int = 5,
        use_graph_context: bool = False,
        normalization: str = "instance",
        feedforward_hidden: int = 512,
        bias: bool = False,
        train_decode_type: str = "sampling",
        val_decode_type: str = "sampling",
        test_decode_type: str = "sampling",
    ):
        super().__init__()
        self.stage_cnt = stage_cnt

        self.encoders: list[MatNetEncoder] = nn.ModuleList(
            [
                MatNetEncoder(
                    embed_dim=embed_dim,
                    num_heads=num_heads,
                    num_layers=num_encoder_layers,
                    normalization=normalization,
                    feedforward_hidden=feedforward_hidden,
                    bias=bias,
                    init_embedding_kwargs={"mode": "RandomOneHot"},
                )
                for _ in range(self.stage_cnt)
            ]
        )
        self.decoders: list[MultiStageFFSPDecoder] = nn.ModuleList(
            [
                MultiStageFFSPDecoder(embed_dim, num_heads, use_graph_context)
                for _ in range(self.stage_cnt)
            ]
        )

        self.train_decode_type = train_decode_type
        self.val_decode_type = val_decode_type
        self.test_decode_type = test_decode_type

    def pre_forward(self, td: TensorDict, env: FFSPEnv, num_starts: int):
        run_time_list = td["run_time"].chunk(env.num_stage, dim=-1)
        for stage_idx in range(self.stage_cnt):
            td["cost_matrix"] = run_time_list[stage_idx]
            encoder = self.encoders[stage_idx]
            embeddings, _ = encoder(td)
            decoder = self.decoders[stage_idx]
            decoder._precompute_cache(embeddings)

        if num_starts > 1:
            # repeat num_start times
            td = batchify(td, num_starts)
            # update machine idx and action mask
            td = env.pre_step(td)

        return td

    def forward(
        self,
        td: TensorDict,
        env: FFSPEnv,
        phase="train",
        num_starts=1,
        return_actions: bool = True,
        **decoder_kwargs,
    ):
        assert not env.flatten_stages, "Multistage model only supports unflattened env"
        assert num_starts <= factorial(env.num_machine)

        # Get decode type depending on phase
        decode_type = getattr(self, f"{phase}_decode_type")
        device = td.device

        td = self.pre_forward(td, env, num_starts)

        # NOTE: this must come after pre_forward due to batchify op
        batch_size = td.size(0)
        logp_list = torch.zeros(size=(batch_size, 0), device=device)
        action_list = []

        while not td["done"].all():
            action_stack = torch.empty(
                size=(batch_size, self.stage_cnt), dtype=torch.long, device=device
            )
            logp_stack = torch.empty(size=(batch_size, self.stage_cnt), device=device)

            for stage_idx in range(self.stage_cnt):
                decoder = self.decoders[stage_idx]
                action, logp = decoder(td, decode_type, num_starts, **decoder_kwargs)
                action_stack[:, stage_idx] = action
                logp_stack[:, stage_idx] = logp

            gathering_index = td["stage_idx"][:, None]
            # shape: (batch, 1)
            action = action_stack.gather(dim=1, index=gathering_index).squeeze(dim=1)
            logp = logp_stack.gather(dim=1, index=gathering_index).squeeze(dim=1)
            # shape: (batch)
            action_list.append(action)
            # transition
            td.set("action", action)
            td = env.step(td)["next"]

            logp_list = torch.cat((logp_list, logp[:, None]), dim=1)

        out = {
            "reward": td["reward"],
            "log_likelihood": logp_list.sum(1),
        }

        if return_actions:
            out["actions"] = torch.stack(action_list, 1)

        return out



import torch


class RewardScaler:
    """This class calculates the running mean and variance of a stepwise observed
    quantity, like the RL reward / advantage using the Welford online algorithm.
    The mean and variance are either used to standardize the input (scale='norm') or
    to scale it (scale='scale').

    Args:
        scale: None | 'scale' | 'mean': specifies how to transform the input; defaults to None
    """

    def __init__(self, scale: str = None):
        self.scale = scale
        self.count = 0
        self.mean = 0
        self.M2 = 0

    def __call__(self, scores: torch.Tensor):
        if self.scale is None:
            return scores
        elif isinstance(self.scale, int):
            return scores / self.scale
        # Score scaling
        self.update(scores)
        tensor_to_kwargs = dict(dtype=scores.dtype, device=scores.device)
        std = (self.M2 / (self.count - 1)).float().sqrt()
        score_scaling_factor = std.to(**tensor_to_kwargs) + torch.finfo(scores.dtype).eps
        if self.scale == "norm":
            scores = (scores - self.mean.to(**tensor_to_kwargs)) / score_scaling_factor
        elif self.scale == "scale":
            scores /= score_scaling_factor
        else:
            raise ValueError("unknown scaling operation requested: %s" % self.scale)
        return scores

    @torch.no_grad()
    def update(self, batch: torch.Tensor):
        batch = batch.reshape(-1)
        self.count += len(batch)

        # newvalues - oldMean
        delta = batch - self.mean
        self.mean += (delta / self.count).sum()
        # newvalues - newMeant
        delta2 = batch - self.mean
        self.M2 += (delta * delta2).sum()



import torch

from tensordict.tensordict import TensorDict

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(
    td: TensorDict, actions=None, ax=None, scale_xy: bool = False, vehicle_capacity=None
):
    import matplotlib.pyplot as plt
    import numpy as np

    from matplotlib import cm, colormaps

    num_routine = (actions == 0).sum().item() + 2
    base = colormaps["nipy_spectral"]
    color_list = base(np.linspace(0, 1, num_routine))
    cmap_name = base.name + str(num_routine)
    out = base.from_list(cmap_name, color_list, num_routine)

    if ax is None:
        _, ax = plt.subplots(dpi=100, figsize=(6, 6))

    td = td.detach().cpu()

    if actions is None:
        actions = td.get("action", None)

    if td.batch_size != torch.Size([]):
        td = td[0]
        actions = actions[0]

    locs = td["locs"]
    scale_demand = td["capacity_original"]
    demands_linehaul = td["demand_linehaul"] * scale_demand
    demands_backhaul = td["demand_backhaul"] * scale_demand
    # scale to closest integer
    demands_linehaul = demands_linehaul.round().int()
    demands_backhaul = demands_backhaul.round().int()

    if actions is None:
        log.warning("No action in TensorDict, rendering unsorted locs")
    else:
        actions = torch.cat([torch.tensor([0]), actions, torch.tensor([0])])

    # Depot
    ax.scatter(
        locs[0, 0],
        locs[0, 1],
        edgecolors=cm.Set2(2),
        facecolors="none",
        s=100,
        linewidths=2,
        marker="s",
        alpha=1,
    )

    for node_idx, loc in enumerate(locs):
        if node_idx == 0:
            continue
        delivery, pickup = demands_linehaul[node_idx], demands_backhaul[node_idx]
        if delivery > 0:
            ax.text(
                loc[0],
                loc[1] + 0.02,
                f"{delivery.item()}",
                horizontalalignment="center",
                verticalalignment="bottom",
                fontsize=10,
                color=cm.Set2(0),
            )
            # scatter delivery as downward triangle
            ax.scatter(
                loc[0],
                loc[1],
                edgecolors=cm.Set2(0),
                facecolors="none",
                s=30,
                linewidths=2,
                marker="v",
                alpha=1,
            )
        elif pickup > 0:
            ax.text(
                loc[0],
                loc[1] - 0.02,
                f"{pickup.item()}",
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=10,
                color=cm.Set2(1),
            )
            ax.scatter(
                loc[0],
                loc[1],
                edgecolors=cm.Set2(1),
                facecolors="none",
                s=30,
                linewidths=2,
                marker="^",
                alpha=1,
            )
        else:
            print("Error: no demand")

    color_idx = 0
    next_actions = torch.roll(actions, -1, 0)
    for ai, aj in zip(actions, next_actions):
        if ai == 0:
            color_idx += 1
        from_loc = locs[ai]
        to_loc = locs[aj]
        # if any of from_loc or to_loc is depot, change color and linewidth
        if ai == 0 or aj == 0:
            color, lw, alpha, style = "lightgrey", 1, 0.5, "--"
        else:
            color, lw, alpha, style = out(color_idx), 1, 1, ""
        ax.plot(
            [from_loc[0], to_loc[0]],
            [from_loc[1], to_loc[1]],
            color=color,
            lw=lw,
            alpha=alpha,
            linestyle=style,
        )
        ax.annotate(
            "",
            xy=(to_loc[0], to_loc[1]),
            xytext=(from_loc[0], from_loc[1]),
            arrowprops=dict(arrowstyle="->", color=color, lw=lw, alpha=alpha),
            size=15,
            annotation_clip=False,
        )

    if scale_xy:
        ax.set_xlim(-0.05, 1.05)
        ax.set_ylim(-0.05, 1.05)

    return ax



import matplotlib.pyplot as plt
import numpy as np
import torch

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(self, decaps, probe, action_mask, ax=None, legend=True):
    """
    Plot a grid of 1x1 squares representing the environment.
    The keepout regions are the action_mask - decaps - probe
    """

    settings = {
        0: {"color": "white", "label": "available"},
        1: {"color": "grey", "label": "keepout"},
        2: {"color": "tab:red", "label": "probe"},
        3: {"color": "tab:blue", "label": "decap"},
    }

    nonzero_indices = torch.nonzero(~action_mask, as_tuple=True)[0]
    keepout = torch.cat([nonzero_indices, probe, decaps.squeeze(-1)])
    unique_elements, counts = torch.unique(keepout, return_counts=True)
    keepout = unique_elements[counts == 1]

    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(6, 6))

    grid = np.meshgrid(np.arange(0, self.size), np.arange(0, self.size))
    grid = np.stack(grid, axis=-1)

    # Add new dimension to grid filled up with 0s
    grid = np.concatenate([grid, np.zeros((self.size, self.size, 1))], axis=-1)

    # Add keepout = 1
    grid[keepout // self.size, keepout % self.size, 2] = 1
    # Add probe = 2
    grid[probe // self.size, probe % self.size, 2] = 2
    # Add decaps = 3
    grid[decaps // self.size, decaps % self.size, 2] = 3

    xdim, ydim = grid.shape[0], grid.shape[1]
    ax.imshow(np.zeros((xdim, ydim)), cmap="gray")

    ax.set_xlim(0, xdim)
    ax.set_ylim(0, ydim)

    for i in range(xdim):
        for j in range(ydim):
            color = settings[grid[i, j, 2]]["color"]
            x, y = grid[i, j, 0], grid[i, j, 1]
            ax.add_patch(plt.Rectangle((x, y), 1, 1, color=color, linestyle="-"))

    # Add grid with 1x1 squares
    ax.grid(which="major", axis="both", linestyle="-", color="k", linewidth=1, alpha=0.5)
    # set 10 ticks
    ax.set_xticks(np.arange(0, xdim, 1))
    ax.set_yticks(np.arange(0, ydim, 1))

    # Invert y axis
    ax.invert_yaxis()

    # Add legend
    if legend:
        num_unique = 4
        handles = [
            plt.Rectangle((0, 0), 1, 1, color=settings[i]["color"])
            for i in range(num_unique)
        ]
        ax.legend(
            handles,
            [settings[i]["label"] for i in range(num_unique)],
            ncol=num_unique,
            loc="upper center",
            bbox_to_anchor=(0.5, 1.1),
        )

    return ax



import matplotlib.pyplot as plt
import numpy as np
import torch

from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


def render(self, td, actions=None, ax=None, legend=True, settings=None):
    """Plot a grid of squares representing the environment.
    The keepout regions are the action_mask - decaps - probe
    """

    from matplotlib.lines import Line2D
    from matplotlib.patches import Annulus, Rectangle, RegularPolygon

    if settings is None:
        settings = {
            "available": {"color": "white", "label": "available"},
            "keepout": {"color": "grey", "label": "keepout"},
            "probe": {"color": "tab:red", "label": "probe"},
            "decap": {"color": "tab:blue", "label": "decap"},
        }

    def draw_capacitor(ax, x, y, color="black"):
        # Backgrund rectangle: same as color but with alpha=0.5
        ax.add_patch(Rectangle((x, y), 1, 1, color=color, alpha=0.5))

        # Create the plates of the capacitor
        plate_width, plate_height = (
            0.3,
            0.1,
        )  # Width and height switched to make vertical
        plate_gap = 0.2
        plate1 = Rectangle(
            (x + 0.5 - plate_width / 2, y + 0.5 - plate_height - plate_gap / 2),
            plate_width,
            plate_height,
            color=color,
        )
        plate2 = Rectangle(
            (x + 0.5 - plate_width / 2, y + 0.5 + plate_gap / 2),
            plate_width,
            plate_height,
            color=color,
        )

        # Add the plates to the axes
        ax.add_patch(plate1)
        ax.add_patch(plate2)

        # Add connection lines (wires)
        line_length = 0.2
        line1 = Line2D(
            [x + 0.5, x + 0.5],
            [
                y + 0.5 - plate_height - plate_gap / 2 - line_length,
                y + 0.5 - plate_height - plate_gap / 2,
            ],
            color=color,
        )
        line2 = Line2D(
            [x + 0.5, x + 0.5],
            [
                y + 0.5 + plate_height + plate_gap / 2,
                y + 0.5 + plate_height + plate_gap / 2 + line_length,
            ],
            color=color,
        )

        # Add the lines to the axes
        ax.add_line(line1)
        ax.add_line(line2)

    def draw_probe(ax, x, y, color="black"):
        # Backgrund rectangle: same as color but with alpha=0.5
        ax.add_patch(Rectangle((x, y), 1, 1, color=color, alpha=0.5))
        ax.add_patch(Annulus((x + 0.5, y + 0.5), (0.2, 0.2), 0.1, color=color))

    def draw_keepout(ax, x, y, color="black"):
        # Backgrund rectangle: same as color but with alpha=0.5
        ax.add_patch(Rectangle((x, y), 1, 1, color=color, alpha=0.5))
        ax.add_patch(
            RegularPolygon((x + 0.5, y + 0.5), numVertices=6, radius=0.45, color=color)
        )

    size = self.size
    td = td.detach().cpu()
    # if batch_size greater than 0 , we need to select the first batch element
    if td.batch_size != torch.Size([]):
        td = td[0]

    if actions is None:
        actions = td.get("action", None)

    # Transform actions from idx to one-hot
    decaps = torch.zeros(size**2)
    decaps.scatter_(0, actions, 1)
    decaps = decaps.reshape(size, size)

    keepout = ~td["action_mask"].reshape(size, size)
    probes = td["probe"].reshape(size, size)

    if ax is None:
        _, ax = plt.subplots(1, 1, figsize=(6, 6))

    grid = np.meshgrid(np.arange(0, size), np.arange(0, size))
    grid = np.stack(grid, axis=-1)

    xdim, ydim = grid.shape[0], grid.shape[1]
    # ax.imshow(np.zeros((xdim, ydim)), cmap="gray")

    ax.set_xlim(0, xdim)
    ax.set_ylim(0, ydim)

    for i in range(xdim):
        for j in range(ydim):
            x, y = grid[i, j, 0], grid[i, j, 1]

            if decaps[i, j] == 1:
                draw_capacitor(ax, x, y, color=settings["decap"]["color"])
            elif probes[i, j] == 1:
                draw_probe(ax, x, y, color=settings["probe"]["color"])
            elif keepout[i, j] == 1:
                draw_keepout(ax, x, y, color=settings["keepout"]["color"])

    ax.grid(which="major", axis="both", linestyle="-", color="k", linewidth=1, alpha=0.5)
    # set 10 ticks
    ax.set_xticks(np.arange(0, xdim, 1))
    ax.set_yticks(np.arange(0, ydim, 1))

    # Invert y axis
    ax.invert_yaxis()

    # # Add legend
    if legend:
        colors = [settings[k]["color"] for k in settings.keys()]
        labels = [settings[k]["label"] for k in settings.keys()]
        handles = [
            plt.Rectangle(
                (0, 0), 1, 1, color=c, edgecolor="k", linestyle="-", linewidth=1
            )
            for c in colors
        ]
        ax.legend(
            handles,
            [label for label in labels],
            ncol=len(colors),
            loc="upper center",
            bbox_to_anchor=(0.5, 1.1),
        )

    return ax



import torch
import torch.nn as nn

from tensordict import TensorDict

from rl4co.envs import RL4COEnvBase, get_env
from rl4co.models.common.improvement.base import ImprovementPolicy
from rl4co.models.zoo.dact.decoder import DACTDecoder
from rl4co.models.zoo.dact.encoder import DACTEncoder
from rl4co.utils.decoding import DecodingStrategy, get_decoding_strategy
from rl4co.utils.pylogger import get_pylogger

log = get_pylogger(__name__)


class DACTPolicy(ImprovementPolicy):
    """
    DACT Policy based on Ma et al. (2021)
    This model first encodes the input graph and current solution using a DACT encoder (:class:`DACTEncoder`)
    and then decodes the 2-opt action (:class:`DACTDecoder`)

    Args:
        embed_dim: Dimension of the node embeddings
        num_encoder_layers: Number of layers in the encoder
        num_heads: Number of heads in the attention layers
        normalization: Normalization type in the attention layers
        feedforward_hidden: Dimension of the hidden layer in the feedforward network
        env_name: Name of the environment used to initialize embeddings
        pos_type: Name of the used positional encoding method (CPE or APE)
        init_embedding: Module to use for the initialization of the embeddings
        pos_embedding: Module to use for the initialization of the positional embeddings
        temperature: Temperature for the softmax
        tanh_clipping: Tanh clipping value (see Bello et al., 2016)
        train_decode_type: Type of decoding to use during training
        val_decode_type: Type of decoding to use during validation
        test_decode_type: Type of decoding to use during testing
    """

    def __init__(
        self,
        embed_dim: int = 64,
        num_encoder_layers: int = 3,
        num_heads: int = 4,
        normalization: str = "layer",
        feedforward_hidden: int = 64,
        env_name: str = "tsp_kopt",
        pos_type: str = "CPE",
        init_embedding: nn.Module = None,
        pos_embedding: nn.Module = None,
        temperature: float = 1.0,
        tanh_clipping: float = 6.0,
        train_decode_type: str = "sampling",
        val_decode_type: str = "sampling",
        test_decode_type: str = "sampling",
    ):
        super(DACTPolicy, self).__init__()

        self.env_name = env_name

        # Encoder and decoder
        self.encoder = DACTEncoder(
            embed_dim=embed_dim,
            init_embedding=init_embedding,
            pos_embedding=pos_embedding,
            env_name=env_name,
            pos_type=pos_type,
            num_heads=num_heads,
            num_layers=num_encoder_layers,
            normalization=normalization,
            feedforward_hidden=feedforward_hidden,
        )

        self.decoder = DACTDecoder(embed_dim=embed_dim, num_heads=num_heads)

        # Decoding strategies
        self.temperature = temperature
        self.tanh_clipping = tanh_clipping
        self.train_decode_type = train_decode_type
        self.val_decode_type = val_decode_type
        self.test_decode_type = test_decode_type

    def forward(
        self,
        td: TensorDict,
        env: str | RL4COEnvBase = None,
        phase: str = "train",
        return_actions: bool = True,
        return_embeds: bool = False,
        only_return_embed: bool = False,
        actions=None,
        **decoding_kwargs,
    ) -> dict:
        """Forward pass of the policy.

        Args:
            td: TensorDict containing the environment state
            env: Environment to use for decoding. If None, the environment is instantiated from `env_name`. Note that
                it is more efficient to pass an already instantiated environment each time for fine-grained control
            phase: Phase of the algorithm (train, val, test)
            return_actions: Whether to return the actions
            actions: Actions to use for evaluating the policy.
                If passed, use these actions instead of sampling from the policy to calculate log likelihood
            decoding_kwargs: Keyword arguments for the decoding strategy. See :class:`rl4co.utils.decoding.DecodingStrategy` for more information.

        Returns:
            out: Dictionary containing the reward, log likelihood, and optionally the actions and entropy
        """

        # Encoder: get encoder output and initial embeddings from initial state
        NFE, PFE = self.encoder(td)
        h_featrues = torch.cat((NFE, PFE), -1)

        if only_return_embed:
            return {"embeds": h_featrues.detach()}

        # Instantiate environment if needed
        if isinstance(env, str) or env is None:
            env_name = self.env_name if env is None else env
            log.info(f"Instantiated environment not provided; instantiating {env_name}")
            env = get_env(env_name)
        assert env.two_opt_mode, "DACT only support 2-opt"

        # Get decode type depending on phase and whether actions are passed for evaluation
        decode_type = decoding_kwargs.pop("decode_type", None)
        if actions is not None:
            decode_type = "evaluate"
        elif decode_type is None:
            decode_type = getattr(self, f"{phase}_decode_type")

        # Setup decoding strategy
        # we pop arguments that are not part of the decoding strategy
        decode_strategy: DecodingStrategy = get_decoding_strategy(
            decode_type,
            temperature=decoding_kwargs.pop("temperature", self.temperature),
            tanh_clipping=decoding_kwargs.pop("tanh_clipping", self.tanh_clipping),
            mask_logits=True,
            improvement_method_mode=True,
            **decoding_kwargs,
        )

        # Perform the decoding
        batch_size, seq_length = td["rec_current"].size()
        logits = self.decoder(td, NFE, PFE).view(batch_size, -1)

        # Get mask
        mask = env.get_mask(td)
        if "action" in td.keys():
            mask[torch.arange(batch_size), td["action"][:, 0], td["action"][:, 1]] = False
            mask[torch.arange(batch_size), td["action"][:, 1], td["action"][:, 0]] = False
        mask = mask.view(batch_size, -1)

        # Get action and log-likelihood
        logprob, action_sampled = decode_strategy.step(
            logits,
            mask,
            action=(
                actions[:, 0] * seq_length + actions[:, 1]
                if actions is not None
                else None
            ),
        )
        action_sampled = action_sampled.unsqueeze(-1)
        if phase == "train":
            log_likelihood = logprob.gather(1, action_sampled)
        else:
            log_likelihood = torch.zeros(batch_size, device=td.device)

        ## return
        DACT_action = torch.cat(
            (
                action_sampled // seq_length,
                action_sampled % seq_length,
            ),
            -1,
        )

        outdict = {"log_likelihood": log_likelihood, "cost_bsf": td["cost_bsf"]}
        td.set("action", DACT_action)

        if return_embeds:
            outdict["embeds"] = h_featrues.detach()

        if return_actions:
            outdict["actions"] = DACT_action

        return outdict
